/***************************************************************************
 * SANE - Scanner Access Now Easy.

   microtek2.c 

   This file (C) 1998 Bernd Schroeder

   This file is part of the SANE package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.

   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.

   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.

   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice.

 ***************************************************************************

   This file implements a SANE backend for the Microtek ScanMaker 630
   and compatible scanners.

   (feedback to:  bernd@aquila.muc.de)

 ***************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>

#include <math.h>

#include <sane/sane.h>
#include <sane/sanei.h>
#include <sane/sanei_config.h>
#include <sane/sanei_scsi.h>
#include <sane/saneopts.h>

#define BACKEND_NAME microtek2
#include <sane/sanei_backend.h>

#include "microtek2.h"



static int md_num_devices = 0;          /* number of devices from config file */
static Microtek2_Device *md_first_dev = NULL;        /* list of known devices */
static Microtek2_Scanner *ms_first_handle = NULL;    /* list of open scanners */
static int ms_dump_clear = 1;
static int ms_dump = 0;                 /* from config file: */
                                        /* 1: inquiry + scanner attributes */
                                        /* 2: + all scsi commands and data */
                                        /* 3: + all scan data */



/***************************************************************************/
/* Dump all we know about scanner to stderr                                */
/***************************************************************************/

static SANE_Status 
dump_attributes(Microtek2_Info *mi)
{
  int i;

  DBG(30, "dump_attributes: mi=%p\n", mi);

  fprintf(stderr, "\n\nScanner attributes from device structure\n");
  fprintf(stderr, "========================================\n");
  fprintf(stderr, "\nScanner ID...\n");
  fprintf(stderr, "~~~~~~~~~~~~~\n");
  fprintf(stderr, "  Vendor Name:        '%s'\n", mi->vendor);
  fprintf(stderr, "  Model Name:         '%s'\n", mi->model);
  fprintf(stderr, "  Revision:           '%s'\n", mi->revision);
	  
  fprintf(stderr, "  Model Code:         0x%02x (", mi->model_code);
  switch(mi->model_code) 
    {
      case 0x80: fprintf(stderr, "Redondo"); break;
      case 0x81: fprintf(stderr, "Aruba"); break;
      case 0x82: fprintf(stderr, "Bali"); break;
      case 0x83: fprintf(stderr, "Washington"); break;
      case 0x84: fprintf(stderr, "Manhattan"); break;
      case 0x85: fprintf(stderr, "TR3"); break;
      case 0x86: fprintf(stderr, "CCP"); break;
      case 0x87: fprintf(stderr, "Scanmaker V"); break;
      case 0x88: fprintf(stderr, "Scanmaker VI"); break;
      case 0x89: fprintf(stderr, "A3-400"); break;
      case 0x8A: fprintf(stderr, "A3-600"); break;
      case 0x8B: fprintf(stderr, "Watt"); break;
      case 0x8C: fprintf(stderr, "TR6"); break;
      case 0x8D: fprintf(stderr, "Tr3 10-bit"); break;
      case 0x8E: fprintf(stderr, "CCB"); break;
      case 0x8F: fprintf(stderr, "Sun Rise"); break;
      case 0x90: fprintf(stderr, "ScanMaker E3 10-bit"); break;
      case 0x91: fprintf(stderr, "X6"); break;
      default:   fprintf(stderr, "Unknown"); break;
    }
  fprintf(stderr, ")\n");
  fprintf(stderr, "  Device Type Code:   0x%02x (%s),\n", 
          mi->device_type,
	  mi->device_type & MI_DEVTYPE_SCANNER ? "Scanner" : "Unknown type");
          
  fprintf(stderr, "  Scanner type:       "); 
  switch (mi->scanner_type) 
    {
      case MI_TYPE_FLATBED: fprintf(stderr, "Flatbed scanner\n");
          break;
      case MI_TYPE_TRANSPARENCY: fprintf(stderr, "Transparency scanner\n");
          break;
      case MI_TYPE_SHEEDFEED: fprintf(stderr, "Sheet feed scanner\n");
          break;
      default: fprintf(stderr, "Unknown\n");
          break;
    }
  
  fprintf(stderr, "  Supported options:  Automatic document feeder: %s\n",
	  mi->option_device & MI_OPTDEV_ADF ? "Yes" : "No");
  fprintf(stderr, "%22sTransparency media adapter: %s\n",
	  " ", mi->option_device & MI_OPTDEV_TMA ? "Yes" : "No");
  fprintf(stderr, "%22sAuto paper detecting: %s\n",
	  " ", mi->option_device & MI_OPTDEV_ADP ? "Yes" : "No");
  fprintf(stderr, "  Scan button:        %s\n", mi->scnbuttn ? "Yes" : "No"); 
         
  
  fprintf(stderr, "\nImaging Capabilities...\n");
  fprintf(stderr, "~~~~~~~~~~~~~~~~~~~~~~~\n");
  fprintf(stderr, "  Color scanner:      %s\n", (mi->color) ? "Yes" : "No");
  fprintf(stderr, "  Number passes:      %d pass%s\n", 
          (mi->onepass) ? 1 : 3,
          (mi->onepass) ? "" : "es");
  fprintf(stderr, "  Resolution:         X-max: %5d dpi\n%22sY-max: %5d dpi\n",
          mi->max_xresolution, " ",mi->max_yresolution);
  fprintf(stderr, "  Geometry:           Geometric width: %5d pts (%2.2f'')\n", 
          mi->geo_width, (float) mi->geo_width / (float) mi->opt_resolution); 
  fprintf(stderr, "%22sGeometric height:%5d pts (%2.2f'')\n", " ",
          mi->geo_height, (float) mi->geo_height / (float) mi->opt_resolution);
  fprintf(stderr, "  Optical resol.  : %5d\n", mi->opt_resolution);

  fprintf(stderr, "  Modes:              Lineart:     %s\n%22sHalftone:     %s\n",
	  (mi->scanmode & MI_HASMODE_LINEART) ? " Yes" : " No", " ",
	  (mi->scanmode & MI_HASMODE_HALFTONE) ? "Yes" : "No");

  fprintf(stderr,"%22sGray:     %s\n%22sColor:     %s\n", " ",
	  (mi->scanmode & MI_HASMODE_GRAY) ? "    Yes" : "    No", " ",
	  (mi->scanmode & MI_HASMODE_COLOR) ? "   Yes" : "   No");

  fprintf(stderr, "  Depths:             Nibble Gray:  %s\n",
	  (mi->depth & MI_HASDEPTH_NIBBLE) ? "Yes" : "No");
  fprintf(stderr, "%22s10-bit-color: %s\n", " ",
	  (mi->depth & MI_HASDEPTH_10) ? "Yes" : "No");
  fprintf(stderr, "%22s12-bit-color: %s\n", " ",
	  (mi->depth & MI_HASDEPTH_12) ? "yes" : "No");
  fprintf(stderr, "  d/l of HT pattern:  %s\n",
          (mi->has_dnldptrn) ? "Yes" : "No");
  fprintf(stderr, "  Builtin HT patt.:   %d\n", mi->grain_slct);
  
  fprintf(stderr, "  LUT capabilities:   ");
  if ( MI_LUTCAP_NONE(mi->lut_cap) ) 
      fprintf(stderr, "None\n"); 
  if ( mi->lut_cap & MI_LUTCAP_256B )
      fprintf(stderr, " 256 bytes\n"); 
  if ( mi->lut_cap & MI_LUTCAP_1024B )
      fprintf(stderr, "1024 bytes\n"); 
  if ( mi->lut_cap & MI_LUTCAP_1024W )
      fprintf(stderr, "1024 words\n"); 
  if ( mi->lut_cap & MI_LUTCAP_4096B )
      fprintf(stderr, "4096 bytes\n"); 
  if ( mi->lut_cap & MI_LUTCAP_4096W )
      fprintf(stderr, "4096 words\n"); 
  
  fprintf(stderr, "\nMiscellaneous capabilities...\n");
  fprintf(stderr, "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
  fprintf(stderr, "  Data format:        ");
  if ( mi->onepass) 
    {
      switch(mi->data_format) 
        {
	  case MI_DATAFMT_CHUNKY:
              fprintf(stderr, "Chunky data, R, G & B in on pixel\n");
              break;
	  case MI_DATAFMT_LPLCONCAT:
              fprintf(stderr, "Line by line in concatenated sequence,\n");
              fprintf(stderr, "%22swithout color indicator\n", " ");
              break;
	  case MI_DATAFMT_LPLSEGREG:
              fprintf(stderr, "Line by line in segregated sequence,\n");
              fprintf(stderr, "%22swith color indicator\n", " ");
              break;
	  case MI_DATAFMT_WORDCHUNKY:
              fprintf(stderr, "Word chunky data\n");
              break;
          default:
              fprintf(stderr, "Unknown\n");
          break;
        }
    } 
  else
      fprintf(stderr, "No information with 3-pass scanners\n");

  fprintf(stderr, "  Color Sequence:     ");
  for ( i = 0; i < RSA_COLORSEQUENCE_L; i++) 
    {
      switch(mi->color_sequence[i]) 
        {
	  case MI_COLSEQ_RED:   fprintf(stderr,"R"); break;
	  case MI_COLSEQ_GREEN: fprintf(stderr,"G"); break;
	  case MI_COLSEQ_BLUE:  fprintf(stderr,"B"); break;
        }
      if ( i == RSA_COLORSEQUENCE_L - 1)
          fprintf(stderr, "\n");
      else          
          fprintf(stderr, " - ");
    }
  fprintf(stderr, "  CCD gap:            %d lines\n", mi->ccd_gap);
  fprintf(stderr, "  CCD pixels:         %d\n", mi->ccd_pixels);
  fprintf(stderr, "  Calib wh str loc:   %d\n", mi->calib_white);
  fprintf(stderr, "  Max calib space:    %d\n", mi->calib_space);
  fprintf(stderr, "  Number of lens:     %d\n", mi->nlens);
  fprintf(stderr, "  Max no of windows:  %d\n", mi->nwindows);
  fprintf(stderr, "  Sh trnsf func equ:  %d\n", mi->shtrnsferequ);
  fprintf(stderr, "  Buffer type:        %s\n", 
          mi->buftype ? "Ping-Pong" : "Ring");
  fprintf(stderr, "  FEPROM:             %s\n", mi->feprom ? "Yes" : "No");

  ms_dump_clear = 0;
  return SANE_STATUS_GOOD;
}
                


static SANE_Status
id_microtek2(Microtek2_Info *mi, SANE_String *model_string)
{
    DBG(15, "id_microtek2:\n");
    if ( strncmp("MICROTEK", mi->vendor, 8) == 0 
      || strncmp("        ", mi->vendor, 8) == 0 ) 
      {
        switch (mi->model_code) 
          {
            case 0x8c :
              *model_string = "ScanMaker 630";
              break;
            default :
              DBG(1, "id_microtek2: Device not a Microtek SCSI-II\n");
              DBG(1, "id_microtek2: Vendor=%s, modelcode=0x%02x\n", 
                      mi->vendor, mi->model_code);
              return SANE_STATUS_INVAL;
          }
      }
    else
      {
        DBG(1, "id_microtek2: Device not a Microtek\n");
        DBG(1, "id_microtek2: Vendor=%s, modelcode=0x%02x\n", 
                mi->vendor, mi->model_code);
        return SANE_STATUS_INVAL;
      }

    return SANE_STATUS_GOOD; 
}


/*---------- sane_cancel() ---------------------------------------------------*/

void
sane_cancel (SANE_Handle handle)
{
    Microtek2_Scanner *ms = handle;

    DBG(30, "sane_cancel: handle=%p\n", handle);
    cancel_scan(ms);
    ms->cancelled = SANE_TRUE;
}


/*---------- sane_close() ----------------------------------------------------*/

void
sane_close (SANE_Handle handle)
{
    Microtek2_Scanner *ms = handle;
    int i;

    DBG(30, "sane_close: ms=%p\n", ms);

    if ( ! ms )
        return;

    /* free malloc'ed and strdup'ed stuff */

    for ( i = 0; i < 3; i++ )
      {
        if ( ms->gamma_table[i] != NULL )
          {
            free((void *) ms->gamma_table[i]);
            ms->gamma_table[i] = NULL;
          }
      }
 
    if ( ms->src_buffer ) 
      {
        free((void *) ms->src_buffer);
        ms->src_buffer = NULL;
      }
    if ( ms->prehold_buf ) 
      {
        free((void *) ms->prehold_buf);
        ms->prehold_buf = NULL;
      }
  
    /* remove Scanner from linked list */
    if ( ms_first_handle == ms )
        ms_first_handle = ms->next;
    else 
      {
        Microtek2_Scanner *ts = ms_first_handle;
        while ( (ts != NULL) && (ts->next != ms) )
            ts = ts->next;
        ts->next = ts->next->next; /* == ms->next */
      }
    
    free(ms);
    ms = NULL;
}


/*---------- sane_control_option() -------------------------------------------*/

SANE_Status
sane_control_option(SANE_Handle handle, SANE_Int option, 
                    SANE_Action action, void *value, SANE_Int *info)
{
    Microtek2_Scanner *ms = handle;
    Microtek2_Device *md;
    Microtek2_Option_Value *val;
    SANE_Option_Descriptor *sod;
    SANE_Status status;


    md = ms->dev;
    val = &ms->val[0];
    sod = &ms->sod[0];

    if ( ms->scanning || ms->this_pass )
        return SANE_STATUS_DEVICE_BUSY;
    
    if ( option < 0 || option > NUM_OPTIONS ) 
      {
        DBG(10, "sane_control_option: option %d invalid\n", option);
        return SANE_STATUS_INVAL;
      }
    
    if ( ! SANE_OPTION_IS_ACTIVE(ms->sod[option].cap) ) 
      {
        DBG(30, "sane_control_option: option %d not active\n", option);
        return SANE_STATUS_INVAL;
      }

    if ( info )
        *info = 0;

    switch ( action ) 
      {
        case SANE_ACTION_GET_VALUE:
          switch ( option ) 
            {
              /* word options */   
              case OPT_RESOLUTION:
              case OPT_Y_RESOLUTION:
              case OPT_THRESHOLD:
              case OPT_TL_X:
              case OPT_TL_Y:
              case OPT_BR_X:
              case OPT_BR_Y:
              case OPT_PREVIEW:
              case OPT_BRIGHTNESS:
              case OPT_CONTRAST:
              case OPT_SHADOW:
              case OPT_SHADOW_R:
              case OPT_SHADOW_G:
              case OPT_SHADOW_B:
              case OPT_MIDTONE:
              case OPT_MIDTONE_R:
              case OPT_MIDTONE_G:
              case OPT_MIDTONE_B:
              case OPT_HIGHLIGHT:
              case OPT_HIGHLIGHT_R:
              case OPT_HIGHLIGHT_G:
              case OPT_HIGHLIGHT_B:
              case OPT_GAMMA_SCALAR:
              case OPT_GAMMA_SCALAR_R:
              case OPT_GAMMA_SCALAR_G:
              case OPT_GAMMA_SCALAR_B:
                *(SANE_Word *) value = val[option].w;
                if (sod[option].type == SANE_TYPE_FIXED )
                    DBG(30, "sane_control_option: opt=%d, act=%d, val=%f\n",
                             option, action, SANE_UNFIX(val[option].w)); 
                else
                    DBG(30, "sane_control_option: opt=%d, act=%d, val=%d\n",
                             option, action, val[option].w); 
                return SANE_STATUS_GOOD;
              
              /* boolean options */
              case OPT_RESOLUTION_BIND:
              case OPT_GAMMA_BIND:
                *(SANE_Bool *) value = val[option].w;
                DBG(30, "sane_control_option: opt=%d, act=%d, val=%d\n",
                         option, action, val[option].w); 
                return SANE_STATUS_GOOD;
              
              /* string options */ 
              case OPT_MODE:
              case OPT_HALFTONE_PATTERN:
              case OPT_CHANNEL:
              case OPT_GAMMA_MODE:
                strcpy(value, val[option].s);
                DBG(30, "sane_control_option: opt=%d, act=%d, val=%s\n",
                         option, action, val[option].s); 
                return SANE_STATUS_GOOD;
              
              /* others */
              case OPT_NUM_OPTS:
                *(SANE_Word *) value = NUM_OPTIONS;
                return SANE_STATUS_GOOD;
 
              default:
                return SANE_STATUS_UNSUPPORTED;
            }
          break;

        case SANE_ACTION_SET_VALUE:
          if ( ! SANE_OPTION_IS_SETTABLE(sod[option].cap) )
              return SANE_STATUS_INVAL;

          status = sanei_constrain_value(ms->sod + option, value, info);
          if (status != SANE_STATUS_GOOD)
              return status;
 
          switch ( sod[option].type )
            {
              case SANE_TYPE_BOOL:
              case SANE_TYPE_INT:
                DBG(40, "sane_control_option: option=%d, action=%d, value=%d\n",
                         option, action, *(SANE_Int *) value);

                if ( val[option].w == *(SANE_Int *) value ) /* no change */
                    return SANE_STATUS_GOOD;
                val[option].w = *(SANE_Int *) value;
                break;
              case SANE_TYPE_FIXED:
                DBG(40, "sane_control_option: option=%d, action=%d, value=%f\n",
                         option, action, SANE_UNFIX( *(SANE_Fixed *) value));
                if ( val[option].w == *(SANE_Fixed *) value ) /* no change */
                    return SANE_STATUS_GOOD;
                val[option].w = *(SANE_Fixed *) value;
                break;
              case SANE_TYPE_STRING:
                DBG(40, "sane_control_option: option=%d, action=%d, value=%s\n",
                         option, action, (SANE_String) value);
                if ( strcmp(val[option].s, (SANE_String) value) == 0 )
                    return SANE_STATUS_GOOD;         /* no change */
                if ( val[option].s )
                    free((void *) val[option].s);
                val[option].s = strdup(value); 
                if ( val[option].s == NULL ) 
                  {
                    DBG(1, "sane_control_option: strdup failed\n");
                    return SANE_STATUS_NO_MEM;
                  }
                break;
              default:
                DBG(1, "sane_control_option: unknown type %d\n",
                        sod[option].type);
                break;
            }

          switch ( option ) 
            {
              case OPT_Y_RESOLUTION:
              case OPT_TL_X:
              case OPT_TL_Y:
              case OPT_BR_X:
              case OPT_BR_Y:
                if ( info )
                    *info |= SANE_INFO_RELOAD_PARAMS;
              case OPT_PREVIEW:
              case OPT_BRIGHTNESS:
              case OPT_THRESHOLD:
              case OPT_CONTRAST:
              case OPT_GAMMA_SCALAR:
              case OPT_GAMMA_SCALAR_R:
              case OPT_GAMMA_SCALAR_G:
              case OPT_GAMMA_SCALAR_B:
              case OPT_HALFTONE_PATTERN:
                return SANE_STATUS_GOOD;

              case OPT_MODE:
                if ( info )
                    *info |= SANE_INFO_RELOAD_PARAMS | SANE_INFO_RELOAD_OPTIONS;

                if ( strcmp(val[option].s, MS_MODESTRING_COLOR24) == 0 
                   || strcmp(val[option].s, MS_MODESTRING_COLOR30) == 0 
                   || strcmp(val[option].s, MS_MODESTRING_COLOR36) == 0 )
                  {
                    /* activate brightness,..., deactivate halftone pattern */
                    /* and threshold */
                    sod[OPT_BRIGHTNESS].cap &= ~SANE_CAP_INACTIVE;
                    sod[OPT_CONTRAST].cap &= ~SANE_CAP_INACTIVE;
                    sod[OPT_CHANNEL].cap &= ~SANE_CAP_INACTIVE;
                    sod[OPT_SHADOW].cap &= ~SANE_CAP_INACTIVE;
                    sod[OPT_MIDTONE].cap &= ~SANE_CAP_INACTIVE;
                    sod[OPT_HIGHLIGHT].cap &= ~SANE_CAP_INACTIVE;
                    sod[OPT_HALFTONE_PATTERN].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_THRESHOLD].cap |= SANE_CAP_INACTIVE;
                    if ( SANE_OPTION_IS_SETTABLE(sod[OPT_GAMMA_MODE].cap) )
                        restore_gamma_options(sod, val, MS_MODE_COLOR);
                  } 
                else if ( strcmp(val[option].s, MS_MODESTRING_GRAY12) == 0 
                        || strcmp(val[option].s, MS_MODESTRING_GRAY10) == 0 
                        || strcmp(val[option].s, MS_MODESTRING_GRAY8) == 0
                        || strcmp(val[option].s, MS_MODESTRING_GRAY2) == 0 )
                  {
                    sod[OPT_BRIGHTNESS].cap &= ~SANE_CAP_INACTIVE;
                    sod[OPT_CONTRAST].cap &= ~SANE_CAP_INACTIVE;
                    sod[OPT_CHANNEL].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_SHADOW].cap &= ~SANE_CAP_INACTIVE;
                    sod[OPT_MIDTONE].cap &= ~SANE_CAP_INACTIVE;
                    sod[OPT_HIGHLIGHT].cap &= ~SANE_CAP_INACTIVE;
                    sod[OPT_HALFTONE_PATTERN].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_THRESHOLD].cap |= SANE_CAP_INACTIVE;
                    if ( SANE_OPTION_IS_SETTABLE(sod[OPT_GAMMA_MODE].cap) )
                        restore_gamma_options(sod, val, MS_MODE_GRAY);
                  }
                else if ( strcmp(val[option].s, MS_MODESTRING_HALFTONE) == 0 )
                  {
                    sod[OPT_BRIGHTNESS].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_CONTRAST].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_CHANNEL].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_SHADOW].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_MIDTONE].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_HIGHLIGHT].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_HALFTONE_PATTERN].cap &= ~SANE_CAP_INACTIVE;
                    sod[OPT_THRESHOLD].cap |= SANE_CAP_INACTIVE;
                    if ( SANE_OPTION_IS_SETTABLE(sod[OPT_GAMMA_MODE].cap) )
                        restore_gamma_options(sod, val, MS_MODE_HALFTONE);
                  }
                else if ( strcmp(val[option].s, MS_MODESTRING_LINEART) == 0 )
                  {
                    sod[OPT_BRIGHTNESS].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_CONTRAST].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_CHANNEL].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_SHADOW].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_MIDTONE].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_HIGHLIGHT].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_HALFTONE_PATTERN].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_THRESHOLD].cap &= ~SANE_CAP_INACTIVE;
                    if ( SANE_OPTION_IS_SETTABLE(sod[OPT_GAMMA_MODE].cap) )
                        restore_gamma_options(sod, val, MS_MODE_LINEART);
                  }
                
                /* these ones are always inactive if the mode changes */
                sod[OPT_SHADOW_R].cap |= SANE_CAP_INACTIVE;
                sod[OPT_SHADOW_G].cap |= SANE_CAP_INACTIVE;
                sod[OPT_SHADOW_B].cap |= SANE_CAP_INACTIVE;
                sod[OPT_MIDTONE_R].cap |= SANE_CAP_INACTIVE;
                sod[OPT_MIDTONE_G].cap |= SANE_CAP_INACTIVE;
                sod[OPT_MIDTONE_B].cap |= SANE_CAP_INACTIVE;
                sod[OPT_HIGHLIGHT_R].cap |= SANE_CAP_INACTIVE;
                sod[OPT_HIGHLIGHT_G].cap |= SANE_CAP_INACTIVE;
                sod[OPT_HIGHLIGHT_B].cap |= SANE_CAP_INACTIVE;
                
                return SANE_STATUS_GOOD;

              case OPT_CHANNEL:
                if ( info )
                    *info |= SANE_INFO_RELOAD_OPTIONS;
                if ( strcmp(val[option].s, MD_CHANNEL_MASTER) == 0 )
                  {
                    sod[OPT_SHADOW].cap &= ~SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE].cap &= ~SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT].cap &= ~SANE_CAP_INACTIVE;   
                    sod[OPT_SHADOW_R].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE_R].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT_R].cap |= SANE_CAP_INACTIVE;   
                    sod[OPT_SHADOW_G].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE_G].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT_G].cap |= SANE_CAP_INACTIVE;   
                    sod[OPT_SHADOW_B].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE_B].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT_B].cap |= SANE_CAP_INACTIVE;   
                  }
                else if ( strcmp(val[option].s, MD_CHANNEL_RED) == 0 )
                  {
                    sod[OPT_SHADOW].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT].cap |= SANE_CAP_INACTIVE;   
                    sod[OPT_SHADOW_R].cap &= ~SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE_R].cap &= ~SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT_R].cap &= ~SANE_CAP_INACTIVE;   
                    sod[OPT_SHADOW_G].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE_G].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT_G].cap |= SANE_CAP_INACTIVE;   
                    sod[OPT_SHADOW_B].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE_B].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT_B].cap |= SANE_CAP_INACTIVE;   
                  }
                else if ( strcmp(val[option].s, MD_CHANNEL_GREEN) == 0 )
                  {
                    sod[OPT_SHADOW].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT].cap |= SANE_CAP_INACTIVE;   
                    sod[OPT_SHADOW_R].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE_R].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT_R].cap |= SANE_CAP_INACTIVE;   
                    sod[OPT_SHADOW_G].cap &= ~SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE_G].cap &= ~SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT_G].cap &= ~SANE_CAP_INACTIVE;   
                    sod[OPT_SHADOW_B].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE_B].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT_B].cap |= SANE_CAP_INACTIVE;   
                  }
                else if ( strcmp(val[option].s, MD_CHANNEL_BLUE) == 0 )
                  {
                    sod[OPT_SHADOW].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT].cap |= SANE_CAP_INACTIVE;   
                    sod[OPT_SHADOW_R].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE_R].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT_R].cap |= SANE_CAP_INACTIVE;   
                    sod[OPT_SHADOW_G].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE_G].cap |= SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT_G].cap |= SANE_CAP_INACTIVE;   
                    sod[OPT_SHADOW_B].cap &= ~SANE_CAP_INACTIVE;           
                    sod[OPT_MIDTONE_B].cap &= ~SANE_CAP_INACTIVE;           
                    sod[OPT_HIGHLIGHT_B].cap &= ~SANE_CAP_INACTIVE;   
                  }
                return SANE_STATUS_GOOD;
              
              case OPT_GAMMA_MODE:
                if ( strcmp(val[OPT_MODE].s, MS_MODESTRING_COLOR24) == 0 
                   || strcmp(val[OPT_MODE].s, MS_MODESTRING_COLOR30) == 0 
                   || strcmp(val[OPT_MODE].s, MS_MODESTRING_COLOR36) == 0 )
                    restore_gamma_options(sod, val, MS_MODE_COLOR);
                else if ( strcmp(val[OPT_MODE].s, MS_MODESTRING_GRAY12) == 0 
                        || strcmp(val[OPT_MODE].s, MS_MODESTRING_GRAY10) == 0 
                        || strcmp(val[OPT_MODE].s, MS_MODESTRING_GRAY8) == 0
                        || strcmp(val[OPT_MODE].s, MS_MODESTRING_GRAY2) == 0 )
                    restore_gamma_options(sod, val, MS_MODE_GRAY);
                else
                    restore_gamma_options(sod, val, MS_MODE_HALFTONE);

                if ( info )
                    *info |= SANE_INFO_RELOAD_OPTIONS;
                return SANE_STATUS_GOOD;

              case OPT_GAMMA_BIND:
                if ( val[option].w )
                  {
                    sod[OPT_GAMMA_SCALAR].cap &= ~SANE_CAP_INACTIVE;
                    sod[OPT_GAMMA_SCALAR_R].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_GAMMA_SCALAR_G].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_GAMMA_SCALAR_B].cap |= SANE_CAP_INACTIVE;
                  }  
                else
                  {
                    sod[OPT_GAMMA_SCALAR].cap |= SANE_CAP_INACTIVE;
                    sod[OPT_GAMMA_SCALAR_R].cap &= ~SANE_CAP_INACTIVE;
                    sod[OPT_GAMMA_SCALAR_G].cap &= ~SANE_CAP_INACTIVE;
                    sod[OPT_GAMMA_SCALAR_B].cap &= ~SANE_CAP_INACTIVE;
                  }

                if ( info )
                    *info |= SANE_INFO_RELOAD_OPTIONS;

                return SANE_STATUS_GOOD;

              /* the following code could be shorter */
              case OPT_SHADOW:
                if ( val[option].w >= val[OPT_MIDTONE].w )
                  {
                    val[OPT_MIDTONE].w = val[OPT_SHADOW].w + 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }
                if ( val[OPT_MIDTONE].w >= val[OPT_HIGHLIGHT].w ) 
                    val[OPT_HIGHLIGHT].w = val[OPT_MIDTONE].w + 1;

                return SANE_STATUS_GOOD;
        
              case OPT_SHADOW_R:
                if ( val[option].w >= val[OPT_MIDTONE_R].w )
                  {
                    val[OPT_MIDTONE_R].w = val[OPT_SHADOW_R].w + 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }
                if ( val[OPT_MIDTONE_R].w >= val[OPT_HIGHLIGHT_R].w ) 
                    val[OPT_HIGHLIGHT_R].w = val[OPT_MIDTONE_R].w + 1;

                return SANE_STATUS_GOOD;

              case OPT_SHADOW_G:
                if ( val[option].w >= val[OPT_MIDTONE_G].w )
                  {
                    val[OPT_MIDTONE_G].w = val[OPT_SHADOW_G].w + 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }
                if ( val[OPT_MIDTONE_G].w >= val[OPT_HIGHLIGHT_G].w ) 
                    val[OPT_HIGHLIGHT_G].w = val[OPT_MIDTONE_G].w + 1;

                return SANE_STATUS_GOOD;

              case OPT_SHADOW_B:
                if ( val[option].w >= val[OPT_MIDTONE_B].w )
                  {
                    val[OPT_MIDTONE_B].w = val[OPT_SHADOW_B].w + 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }
                if ( val[OPT_MIDTONE_B].w >= val[OPT_HIGHLIGHT_B].w ) 
                    val[OPT_HIGHLIGHT_B].w = val[OPT_MIDTONE_B].w + 1;

                return SANE_STATUS_GOOD;

              case OPT_MIDTONE:
                if ( val[option].w <= val[OPT_SHADOW].w )
                  {
                    val[OPT_SHADOW].w = val[OPT_MIDTONE].w - 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }
                if ( val[option].w >= val[OPT_HIGHLIGHT].w ) 
                  {
                    val[OPT_HIGHLIGHT].w = val[OPT_MIDTONE].w + 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }

                return SANE_STATUS_GOOD;
        
              case OPT_MIDTONE_R:
                if ( val[option].w <= val[OPT_SHADOW_R].w )
                  {
                    val[OPT_SHADOW_R].w = val[OPT_MIDTONE_R].w - 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }
                if ( val[option].w >= val[OPT_HIGHLIGHT_R].w ) 
                  {
                    val[OPT_HIGHLIGHT_R].w = val[OPT_MIDTONE_R].w + 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }

                return SANE_STATUS_GOOD;

              case OPT_MIDTONE_G:
                if ( val[option].w <= val[OPT_SHADOW_G].w )
                  {
                    val[OPT_SHADOW_G].w = val[OPT_MIDTONE_G].w - 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }
                if ( val[option].w >= val[OPT_HIGHLIGHT_G].w ) 
                  {
                    val[OPT_HIGHLIGHT_G].w = val[OPT_MIDTONE_G].w + 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }

                return SANE_STATUS_GOOD;
        
              case OPT_MIDTONE_B:
                if ( val[option].w <= val[OPT_SHADOW_B].w )
                  {
                    val[OPT_SHADOW_B].w = val[OPT_MIDTONE_B].w - 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }
                if ( val[option].w >= val[OPT_HIGHLIGHT_B].w ) 
                  {
                    val[OPT_HIGHLIGHT_B].w = val[OPT_MIDTONE_B].w + 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }

                return SANE_STATUS_GOOD;
        
              case OPT_HIGHLIGHT:
                if ( val[option].w <= val[OPT_MIDTONE].w )
                  {
                    val[OPT_MIDTONE].w = val[OPT_HIGHLIGHT].w - 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }
                if ( val[OPT_MIDTONE].w <= val[OPT_SHADOW].w ) 
                    val[OPT_SHADOW].w = val[OPT_MIDTONE].w - 1;

                return SANE_STATUS_GOOD;

              case OPT_HIGHLIGHT_R:
                if ( val[option].w <= val[OPT_MIDTONE_R].w )
                  {
                    val[OPT_MIDTONE_R].w = val[OPT_HIGHLIGHT_R].w - 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }
                if ( val[OPT_MIDTONE_R].w <= val[OPT_SHADOW_R].w ) 
                    val[OPT_SHADOW_R].w = val[OPT_MIDTONE_R].w - 1;

                return SANE_STATUS_GOOD;

              case OPT_HIGHLIGHT_G:
                if ( val[option].w <= val[OPT_MIDTONE_G].w )
                  {
                    val[OPT_MIDTONE_G].w = val[OPT_HIGHLIGHT_G].w - 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }
                if ( val[OPT_MIDTONE_G].w <= val[OPT_SHADOW_G].w ) 
                    val[OPT_SHADOW_G].w = val[OPT_MIDTONE_G].w - 1;

                return SANE_STATUS_GOOD;

              case OPT_HIGHLIGHT_B:
                if ( val[option].w <= val[OPT_MIDTONE_B].w )
                  {
                    val[OPT_MIDTONE_B].w = val[OPT_HIGHLIGHT_B].w - 1;
                    if ( info )
                        *info |= SANE_INFO_RELOAD_OPTIONS;
                  }
                if ( val[OPT_MIDTONE_B].w <= val[OPT_SHADOW_B].w ) 
                    val[OPT_SHADOW_B].w = val[OPT_MIDTONE_B].w - 1;

                return SANE_STATUS_GOOD;

              case OPT_RESOLUTION:
                if ( ms->val[OPT_RESOLUTION_BIND].w == SANE_TRUE )
                    ms->val[OPT_Y_RESOLUTION].w = ms->val[OPT_RESOLUTION].w;
                if ( info )
                    *info |= SANE_INFO_RELOAD_PARAMS;
                return SANE_STATUS_GOOD;

              case OPT_RESOLUTION_BIND:
                if ( ms->val[option].w == SANE_FALSE )
                  {
                    ms->sod[OPT_Y_RESOLUTION].cap &= ~SANE_CAP_INACTIVE;
                    ms->sod[OPT_RESOLUTION].title = "X-Resolution";
                  }
                else
                  {
                    ms->sod[OPT_Y_RESOLUTION].cap |= SANE_CAP_INACTIVE;
                    ms->sod[OPT_RESOLUTION].title = SANE_TITLE_SCAN_RESOLUTION;
                    ms->val[OPT_Y_RESOLUTION].w = ms->val[OPT_RESOLUTION].w;
                  }
                if ( info )
                    *info |= SANE_INFO_RELOAD_OPTIONS;
                return SANE_STATUS_GOOD;
                    
              default:
                return SANE_STATUS_UNSUPPORTED;
            } 
          break; 
        default:
          DBG(1, "sane_control_option: Unsupported action %d\n", action);
          return SANE_STATUS_UNSUPPORTED;
      }
}


/*---------- sane_exit() -----------------------------------------------------*/

void
sane_exit (void)
{
    Microtek2_Device *next;

    DBG(30, "sane_exit:\n");

    /* close all leftover Scanners */
    while (ms_first_handle != NULL)
        sane_close(ms_first_handle);
    /* free up device list */
    while (md_first_dev != NULL) 
      {
        next = md_first_dev->next;
        if ( md_first_dev->halftone_mode_list )
          {
            free((void *) md_first_dev->halftone_mode_list);
            md_first_dev->halftone_mode_list = NULL;
          }
          
        free((void *) md_first_dev);
        md_first_dev = NULL;
        
        md_first_dev = next;
      }
    sane_get_devices(NULL, SANE_FALSE);     /* free list of SANE_Devices */

    DBG(30, "sane_exit: MICROTEK2 says goodbye.\n");
}


/*---------- sane_get_devices()-----------------------------------------------*/

SANE_Status
sane_get_devices(const SANE_Device ***device_list, SANE_Bool local_only)
{
    /* return a list of available devices; available here means that we get */
    /* a positive response to an 'INQUIRY' and possibly to a */
    /* 'READ Scanner Attribute' call */
    
    static const SANE_Device **sd_list = NULL;
    Microtek2_Device *md;
    SANE_Status status;
    int index;

    DBG(30, "sane_get_devices: local_only=%d\n", local_only);
 
    /* first free old list, if there is one; frontend wants a new list */
    if ( sd_list ) 
      {
        free(sd_list);                            /* free array of pointers */ 
        if ( device_list == NULL ) 
          {   
            DBG(30, "sane_get_devices: sd_list_freed\n"); /* called from */
            return SANE_STATUS_GOOD;                      /* sane_exit() */ 
          }
      }

    sd_list = (const SANE_Device **) 
               malloc( (md_num_devices + 1) * sizeof(SANE_Device **));
    if ( ! sd_list ) 
      {
        DBG(5, "sane_get_devices: malloc() for sd_list failed\n");
        return SANE_STATUS_NO_MEM;
      } 
   
    *device_list = sd_list; 
    index = 0;    
    md = md_first_dev;
    while ( md ) 
      {
	status = attach_device(md);
        if ( status != SANE_STATUS_GOOD ) 
          {
            DBG(10, "sane_get_devices: attach_device status '%s'\n", 
                     sane_strstatus(status));
            md = md->next;
            continue;
          }
       
        /* check whether unit is ready, if so add it to the list */
        status = scsi_test_unit_ready(md);
        if ( status != SANE_STATUS_GOOD ) 
          {
            md = md->next;                         
            continue;
          }

        /* md->sane.name = md->name;
        md->sane.vendor = "Microtek";
        md->sane.model = md->info.model; 
        md->sane.type = "flatbed scanner"; */

        sd_list[index] = &md->sane;
        
        ++index;
        md = md->next;                         
      }

    sd_list[index] = NULL;
    return SANE_STATUS_GOOD;
}



/*---------- sane_get_option_descriptor() ------------------------------------*/

const SANE_Option_Descriptor *
sane_get_option_descriptor(SANE_Handle handle, SANE_Int n)
{
    Microtek2_Scanner *ms = handle;


    DBG(30, "sane_get_option_descriptor: handle=%p, opt=%d\n", handle, n);
             
    if ( n < 0 || n > NUM_OPTIONS ) 
      {
        DBG(30, "sane_get_option_descriptor: invalid option %d\n", n);
        return NULL;
      }

    return &ms->sod[n];
}


/*---------- sane_get_parameters() -------------------------------------------*/

SANE_Status
sane_get_parameters(SANE_Handle handle, SANE_Parameters *params)
{
    Microtek2_Scanner *ms = handle;
    Microtek2_Option_Value *val;
    int mode;  
    SANE_Int depth;  
    SANE_Int depth_sane;  
    double x_pixel_per_mm;
    double y_pixel_per_mm;
    double x1_pixel;
    double y1_pixel;
    double width_pixel;
    double height_pixel;
    SANE_Int bytes_per_line;


    DBG(30, "sane_get_parameters: handle=%p, params=%p\n", handle, params);
             

    val= ms->val;

    if ( ! ms->scanning )            /* get an estimate for the params */
      {
        get_scan_mode_and_depth(ms, &mode, &depth);
   
        switch ( mode ) 
          {
	    case MS_MODE_COLOR:
              if ( ms->dev->info.onepass ) 
                {
                  ms->params.format = SANE_FRAME_RGB;
                  ms->params.last_frame = SANE_TRUE;
                } 
              else 
                {   
                  ms->params.format = SANE_FRAME_RED;  /* ?? */
                  ms->params.last_frame = SANE_FALSE;
                }
              break;
	    case MS_MODE_GRAY:
	    case MS_MODE_HALFTONE:
	    case MS_MODE_LINEART:
              ms->params.format = SANE_FRAME_GRAY;
              ms->params.last_frame = SANE_TRUE;
              break;
            default:
              DBG(1, "sane_get_parameters: Unknown scan mode %d\n", mode);
              break;
          } 

      /* calculate depth, SANE has only 1, 8, 16 bits per pixel; */
      /* assume that other bitdepths are scaled into a larger one */ 
      if ( depth == 1 )
          depth_sane = 1;
      else 
          depth_sane = 8 * ((depth + 7) / 8);
    
      ms->params.depth = (SANE_Int) depth_sane;

      /* calculate lines, pixels per line and bytes per line */
      x_pixel_per_mm = SANE_UNFIX(val[OPT_RESOLUTION].w) / MM_PER_INCH;
      y_pixel_per_mm = SANE_UNFIX(val[OPT_Y_RESOLUTION].w) / MM_PER_INCH;
      DBG(30, "sane_get_parameters: y_res=%f, mm_per_inch=%f, ppm=%f\n", 
              SANE_UNFIX(val[OPT_Y_RESOLUTION].w), MM_PER_INCH, y_pixel_per_mm);
      
       
      y1_pixel = SANE_UNFIX(ms->val[OPT_TL_Y].w) * y_pixel_per_mm;
      height_pixel = fabs(SANE_UNFIX(ms->val[OPT_BR_Y].w) * y_pixel_per_mm 
                          - y1_pixel) + 0.5;
      ms->params.lines = (SANE_Int) height_pixel;

      x1_pixel =  SANE_UNFIX(ms->val[OPT_TL_X].w) * x_pixel_per_mm;
      width_pixel = fabs(SANE_UNFIX(ms->val[OPT_BR_X].w) * x_pixel_per_mm 
                         - x1_pixel) + 0.5;
      ms->params.pixels_per_line = (SANE_Int) width_pixel;
                                     
 
      if ( depth_sane == 1 )
          bytes_per_line =  (width_pixel + 7 ) / 8;
      else 
        {
          bytes_per_line = width_pixel * (depth_sane / 8 );
          if ( mode == MS_MODE_COLOR && ms->dev->info.onepass )
	      bytes_per_line *= 3;
        }
      ms->params.bytes_per_line = (SANE_Int) bytes_per_line;
    }  /* if ms->scanning */
  
  if ( params )
     *params = ms->params;     
   
  DBG(30,"sane_get_parameters: format=%d, last_frame=%d, lines=%d\n",
        ms->params.format,ms->params.last_frame, ms->params.lines);
  DBG(30,"sane_get_parameters: depth=%d, ppl=%d, bpl=%d\n",
        ms->params.depth,ms->params.pixels_per_line, ms->params.bytes_per_line);

  return SANE_STATUS_GOOD;
}



SANE_Status
sane_get_select_fd (SANE_Handle handle, SANE_Int * fd)
{
  DBG(10, "sane_get_select_fd\n");
  return SANE_STATUS_UNSUPPORTED;
}




/*---------- sane_init() -----------------------------------------------------*/

SANE_Status
sane_init(SANE_Int *version_code, SANE_Auth_Callback authorize)
{
    FILE *fp;
    char dev_name[PATH_MAX];
    Microtek2_Device *md;


    DBG_INIT();
    DBG(1, "sane_init: Microtek2 (v%d.%d) says hello...\n", 
           MICROTEK2_MAJOR, MICROTEK2_MINOR);

    if ( version_code )
        *version_code = SANE_VERSION_CODE(V_MAJOR, V_MINOR, 0);

    fp = sanei_config_open(MICROTEK2_CONFIG_FILE);
    if ( fp == NULL )
        DBG(10, "sane_init: file not opened: '%s'\n", MICROTEK2_CONFIG_FILE);
    else
      { 
        while ( fgets(dev_name, sizeof(dev_name), fp) ) 
          {
	    add_device_list(dev_name, &md);
            if ( md )
	        attach_device(md);
          }
          fclose(fp);
      }

    if ( md_first_dev == NULL )
      {
        /* config file not found or no valid entry; default to /dev/scanner */
        /* instead of insisting in config file */
	add_device_list("/dev/scanner", &md);
        if ( md )
	    attach_device(md);
      }
    return(SANE_STATUS_GOOD);
}


/*---------- sane_open() -----------------------------------------------------*/

SANE_Status
sane_open(SANE_String_Const name, SANE_Handle *handle)
{
    Microtek2_Scanner *ms;
    Microtek2_Device *md;
    Microtek2_Info *mi;
    SANE_Status status;
    int i;
    
    DBG(30, "sane_open: device='%s'\n", name);

    *handle = NULL;
    md = md_first_dev;
    
    if ( name ) 
      {
	/* add_device_list() returns a pointer to the device struct if */
        /* the device is known or newly added, else it returns NULL */
        
	status = add_device_list(name, &md);
        if ( status != SANE_STATUS_GOOD ) 
            return status;  
      }

    if ( ! md ) 
      {
        DBG(10, "sane_open: invalid device name '%s'\n", name);
        return SANE_STATUS_INVAL;  
      }
  
    /* attach_device executes INQUIRY and READ SCANNER ATTRIBUTES */
    status = attach_device(md);
    if ( status != SANE_STATUS_GOOD ) 
        return status;  
    
    ms = malloc(sizeof(Microtek2_Scanner));
    if ( ms == NULL ) 
      {
        DBG(1, "sane_open: malloc() for ms failed\n");
        return SANE_STATUS_NO_MEM;  
      }
    
    memset(ms, 0, sizeof(ms));
    ms->dev = md;
    mi = &md->info;

    /* Assume that all three colors are transferred to the scanner with one */
    /* SEND GAMMA command, if it is a onepass scanner or a three pass */
    /* scanner and the scan mode is gray. If it is a three pass scanner */
    /* and the scan mode is color, one color per pass is transferred */
    /* I don't know if this is correct. We can use the malloced memory area */
    /* as the data part of the SCSI command if we reserve some room at */
    /* the beginning of the area for then command itself */

    get_lut_size(mi, &md->max_lut_size, &md->lut_entry_size);
    if ( md->max_lut_size > 0 )
      {
        if ( ! mi->onepass && ms->mode == MS_MODE_COLOR )
          {
            for ( i = 0; i < 3; i++ )
              {
                ms->gamma_table[i] = 
                    (u_int8_t *) malloc(md->max_lut_size * md->lut_entry_size
                                      + SG_CMD_L);
                if ( ms->gamma_table[i] == NULL )
                  {
                    DBG(1, "sane_open: malloc for gammatable failed\n");
                    return SANE_STATUS_NO_MEM;
                  }
              }
          }
        else
          {
            ms->gamma_table[0] = 
                (u_int8_t *) malloc( 3 * md->max_lut_size * md->lut_entry_size
                                      + SG_CMD_L);
            if ( ms->gamma_table[0] == NULL )
              {
                DBG(1, "sane_open: malloc for gammatable failed\n");
                return SANE_STATUS_NO_MEM;
              }
          }
      }

    ms->scanning = SANE_FALSE;
    ms->cancelled = SANE_FALSE;
    ms->do_three_pass = (md->info.onepass) ? SANE_FALSE : SANE_TRUE;
    ms->this_pass = 0;
    ms->sfd = -1;
    ms->src_buffer = NULL;
    init_options(ms);
    
    /* insert scanner into linked list */
    ms->next = ms_first_handle;
    ms_first_handle = ms;
 
    *handle = ms;

    return SANE_STATUS_GOOD;
}


/*---------- sane_read() -----------------------------------------------------*/

SANE_Status
sane_read(SANE_Handle handle, SANE_Byte *buf, SANE_Int maxlen, SANE_Int *len )
{
    Microtek2_Scanner *ms = handle;
    Microtek2_Info *mi;
    SANE_Status status;
    SANE_Int dest_max_lines;       /* how many lines fit into the */
                                   /* frontend's buffer */ 


    DBG(30, "sane_read(): handle=%p, buf=%p, maxlen=%d\n", 
             handle, buf, maxlen);

    *len = 0;               /* initialize for status != GOOD */
    
    mi = &ms->dev->info;
    if ( ! ms->scanning ) 
      {
        DBG(15, "sane_read: Scanner %p not scanning\n", ms);
        return SANE_STATUS_INVAL;
      }

    if ( (ms->src_remaining_lines <= 0 && ms->undelivered_lines <= 0)
         || ms->cancelled )
      {
        if ( ms->cancelled )
          {
            status = cancel_scan(ms); 
            DBG(30, "sane_read: Scan cancelled, status '%s'\n",
                 sane_strstatus(status));
            return status;
          }
        else 
            return eof_scan(ms); 
      }


    /* calculate number of lines to read */
    /* TODO: bitdepth > 8, destination buffer shorter than one line */
    do
      {
        dest_max_lines = maxlen / ms->bpl;
        if (dest_max_lines == 0 ) 
          {
            DBG(15, "sane_read: FE buffer is less than one line\n");
            return SANE_STATUS_IO_ERROR;
          }

        ms->src_lines_to_read = MIN(dest_max_lines, 
                               MIN(ms->src_remaining_lines, ms->src_max_lines));
        ms->transfer_length = ms->src_lines_to_read * ms->bpl;
        DBG(30, "sane_read: transferlength=%d, lines=%d linelength=%d\n",
                 ms->transfer_length, ms->src_lines_to_read, ms->bpl);
   
        if ( ms->transfer_length != 0 )
          { 
            status = scsi_wait_for_ready(ms);
            if ( status != SANE_STATUS_GOOD ) 
                return status;
   
            status = scsi_read_image(ms);
            if ( status != SANE_STATUS_GOOD ) 
                return status;
              
            ms->src_remaining_lines -= ms->src_lines_to_read;;
          }

        if ( ms_dump >= 3 )
            dump_area2(ms->src_buffer, ms->transfer_length, "scandata");

        /* test, output color indicator */
        /* for ( i = 0 ; i < ms->transfer_length; i = i + ms->bpl)
            fprintf(stderr,"'%c' '%c' '%c'\n", *(ms->src_buffer + i),
                    *(ms->src_buffer + i + ms->bpl / 3),
                    *(ms->src_buffer + i + ms->bpl / 3 * 2));
        */

        /* prepare data for frontend */
        switch (ms->mode) 
          {
	    case MS_MODE_COLOR:
              if (  ms->do_three_pass )
                /* TODO */
                {
                  DBG(1, "sane_read: 3 pass not yet supported\n");
                  return SANE_STATUS_IO_ERROR;
                }
              else 
                {
                  switch ( ms->dev->info.data_format ) 
                    {
		      case MI_DATAFMT_CHUNKY:
                       /* TODO */
                        DBG(1, "sane_read: Data format %d\n", 
                                 ms->dev->info.data_format);
                        return SANE_STATUS_IO_ERROR;
		      case MI_DATAFMT_LPLCONCAT:
                        /* TODO */
                        DBG(1, "sane_read: Data format %d\n", 
                                ms->dev->info.data_format);
                        return SANE_STATUS_IO_ERROR;
		      case MI_DATAFMT_LPLSEGREG:
		        proc_segreg_data(ms, buf, maxlen, len);
                        break;
		      case MI_DATAFMT_WORDCHUNKY:
                        /* TODO */
                        DBG(1, "sane_read: Data format %d\n", 
                                 ms->dev->info.data_format);
                        return SANE_STATUS_IO_ERROR;
                      default:
                        DBG(1, "sane_read: Unknown data format %d\n", 
                                ms->dev->info.data_format);
                        return SANE_STATUS_IO_ERROR;
                    }      
                }
              break;   
	    case MS_MODE_GRAY:
	      proc_gray_data(ms, buf, len);
              break;
	    case MS_MODE_HALFTONE:
	    case MS_MODE_LINEART:
              proc_onebit_data(ms, buf, len);
              break;
            default:
              DBG(1, "sane_read: Unknown scan mode %d found\n", ms->mode);
              return SANE_STATUS_IO_ERROR;
              break;
          }
      } while ( *len == 0 );
    DBG(30, "sane_read: *len=%d\n", *len);
    return SANE_STATUS_GOOD;
}


/*---------- sane_set_io_mode() ---------------------------------------------*/

SANE_Status
sane_set_io_mode (SANE_Handle handle, SANE_Bool non_blocking)
{
  DBG(30, "sane_set_io_mode:\n");
  return SANE_STATUS_UNSUPPORTED;
}


/*---------- sane_start() ----------------------------------------------------*/

SANE_Status
sane_start(SANE_Handle handle)
{
    Microtek2_Scanner *ms = handle;
    Microtek2_Device *md;
    Microtek2_Info *mi;
    SANE_Status status;
    u_int8_t *pos;
    int lut_size;
    int word_size;
    int color;
    int i;

    DBG(30, "sane_start: handle=0x%p\n", handle);
   
    /* make sure we have an actual parameter set */
    /* sane_get_parameters(ms, NULL);   */
    
    md = ms->dev;
    mi = &ms->dev->info;

    if (ms->sfd < 0) {      /* first or only pass of this scan */

        status = sanei_scsi_open (md->sane.name, &ms->sfd, 
                                  scsi_sense_handler, 0);
        if (status != SANE_STATUS_GOOD) {
            DBG(1, "sane_start: sanei_scsi_open failed: '%s'\n",
                    sane_strstatus(status)); 
            return status;
        }

        /* calculate the parameters from the option values as needed by */
        /* the SET WINDOW command */
        status = get_scan_parameters(ms);
        if ( status != SANE_STATUS_GOOD )
            return status;

        status = scsi_set_window(ms ,1);
        if ( status != SANE_STATUS_GOOD )
            return status;

        status = scsi_read_image_info(ms);
        if ( status != SANE_STATUS_GOOD )
            return status;
        
        /* for ( i = 0; i < 4; i++ ) {
            if ( (status = scsi_read_gamma(ms, i)) != SANE_STATUS_GOOD ) {
                DBG(5, "sane_start: scsi_read_gamma failed: '%s'\n",
                        sane_strstatus(status)); 
                return status;
            }
        } */

        word_size = md->lut_entry_size;
        if ( ! mi->onepass && ms->mode == MS_MODE_COLOR )
          {
            pos = ms->gamma_table[0 /* insert pass */];
            lut_size = md->max_lut_size;
            color = 0;   /* take pass into acount */
          }
        else
          {
            pos = ms->gamma_table[0];
            lut_size =  3 * md->max_lut_size;
            color = 3;
          }

        status = scsi_send_gamma(ms->sfd, pos, lut_size, word_size, color);
        if ( status != SANE_STATUS_GOOD )
            return status;


        /* TODO: format, lastframe, if not onepass, other things with 3 pass */
        /* ms->params.format = ?? ;  */
        /* ms->params.last_frame = ?? ;  */
        ms->params.lines = ms->src_remaining_lines;
        ms->params.depth = ms->depth;
        ms->params.pixels_per_line = ms->ppl;
        /* TODO: calculate *exact* bytes (color indicator, padding byte,..) */
        ms->params.bytes_per_line = ms->bpl;
        
        ms->scanning = SANE_TRUE;
        ms->cancelled = SANE_FALSE;
        ms->undelivered_lines = 0;

        /* calculate number of lines that fit into the scsi buffer */
        ms->src_max_lines = sanei_scsi_max_request_size / ms->bpl;
       
        ms->src_buffer = malloc(sanei_scsi_max_request_size);
        if ( ms->src_buffer == NULL ) {
            DBG(1, "sane_start: malloc for scan buffer failed\n");
            return SANE_STATUS_NO_MEM;
        }

        if ( ms->mode == MS_MODE_COLOR 
             && mi->data_format == MI_DATAFMT_LPLSEGREG ) 
          {             
            /* In this case the data is not neccessarily in the order RGB */
            /* and there may be different numbers of read red, green an blue */
            /* segments. This buffer holds umcomplete lines. Experiments */
            /* have revealed that the number of segments that must be kept */
            /* depends on the y-resolution and is a maximum of 32 segments. */
            /* The following buffer size is determined heuristically. */
            /* Reserve so much space that word transfers fit into the buffer */
            /* I don't know if that is neccessary */ 
            ms->size_of_prehold_buf = mi->geo_width * mi->ccd_gap * 2
                                    * ((int) (SANE_UNFIX(ms->y_resolution_dpi)
                                      / (double) mi->opt_resolution) + 1) 
                                    * 3 * 2;
            DBG(30, "sane_start: preholdbufsize=%d\n", ms->size_of_prehold_buf);
            ms->prehold_buf = malloc(ms->size_of_prehold_buf);
            if ( ms->prehold_buf == NULL )
              {
                DBG(1, "sane_start: malloc for prehold buffer failed\n");
                return SANE_STATUS_NO_MEM;
              }
            for ( i = 0; i < 3; i++ )
              ms->prehold[i] = 0;
          }
        else
          ms->prehold_buf = NULL;
        

        if ( (status = scsi_wait_for_ready(ms)) != SANE_STATUS_GOOD )
            return status;
    }

    return SANE_STATUS_GOOD;

}


/*---------- add_device_list() -----------------------------------------------*/

static SANE_Status
add_device_list(SANE_String_Const dev_name, Microtek2_Device **mdev)
{
    Microtek2_Device *md;
    SANE_String hdev;
    size_t len;


    DBG(30, "add_device_list: device='%s'", dev_name);

    if ( (hdev = strdup(dev_name)) == NULL) { 
	DBG(5, "add_device_list: malloc() for hdev failed\n");
        return SANE_STATUS_NO_MEM;
    }

    len = strlen(dev_name);
    if ( hdev[len - 1] == '\n' )
      {
        --len;
        hdev[len] = '\0';
      }

    if ( hdev[0] == '#' || hdev[0] == '\0' ) 
      {
	DBG(30, "add_device_list: Comment or empty line in %s\n",
                 MICROTEK2_CONFIG_FILE);
        *mdev = NULL;
        return SANE_STATUS_GOOD;       /* ignore empty lines and comments */
      }
    else if ( strncmp(hdev, "dump=1", len) == 0 )
      {
        ms_dump = 1;
        *mdev = NULL;
        return SANE_STATUS_GOOD;
      }
    else if ( strncmp(hdev, "dump=2", len) == 0 )
      {
        ms_dump = 2;
        *mdev = NULL;
        return SANE_STATUS_GOOD;
      }
    else if ( strncmp(hdev, "dump=3", len) == 0 )
      {
        ms_dump = 3;
        *mdev = NULL;
        return SANE_STATUS_GOOD;
      }

    /* check, if it is not already known */
    md = md_first_dev;
    while ( md ) {
        if ( strcmp(hdev, md->name) == 0 ) {
	    DBG(30, "add_device_list: device %s already in list\n", hdev);

            *mdev = md;
            return SANE_STATUS_GOOD;
        }
        md = md->next;
    }

    md = (Microtek2_Device *) malloc(sizeof(Microtek2_Device));
    if ( md == NULL ) {
	DBG(1, "add_device_list: malloc() for md failed\n");
        return SANE_STATUS_NO_MEM;
    }
        
    /* initialize Device and add it at the beginning of the list */
    md->next = md_first_dev;
    md_first_dev = md;
    memset(&md->info, 0, sizeof(Microtek2_Info));
    md->sane.name = NULL;
    md->sane.vendor = NULL;
    md->sane.model = NULL;
    md->sane.type = NULL;
    strncpy(md->name, hdev, PATH_MAX - 1);
    ++md_num_devices;
    *mdev = md;
    free(hdev);

    return SANE_STATUS_GOOD;
}


/*---------- attach_device() -------------------------------------------------*/

static SANE_Status
attach_device(Microtek2_Device *md)
{
    /* This function is called from sane_init() to do the inquiry and to read */
    /* the scanner attributes. If one of these calls fails, or if a new */
    /* device is passed in sane_open() this function may also be called */
    /* from sane_open() or sane_get_devices(). */

    SANE_String model_string; 
    int status;

    
    DBG(30, "attach_device: device='%s'\n", md->name);

    status = scsi_inquiry(&(md->info), md->name);
    if ( status != SANE_STATUS_GOOD ) 
        return status;     

    status = scsi_read_attributes(&(md->info), md->name);
    if ( status != SANE_STATUS_GOOD )
        return status;
    
    if ( id_microtek2(&md->info, &model_string) != SANE_STATUS_GOOD)
        return SANE_STATUS_INVAL;

    md->sane.name = md->name;
    md->sane.vendor = "Microtek";
    md->sane.model = strdup(model_string); 
    md->sane.type = "flatbed scanner";
    
    return SANE_STATUS_GOOD;

}


/*---------- calculate_gamma() -----------------------------------------------*/

static SANE_Status
calculate_gamma(Microtek2_Scanner *ms, u_int8_t *pos, double gamma)
{
    Microtek2_Device *md;
    Microtek2_Info *mi;
    double exp;
    unsigned val;
    double mult;
    double steps;
    int factor;
    int i;
    

    DBG(30, "calculate_gamma: ms=%p, pos=%p\n", ms, pos);

    md = ms->dev;
    mi = &md->info;

    /* does this work everywhere ? */
    if ( mi->depth & MI_HASDEPTH_12 )
        factor = md->max_lut_size / 4096;
    else if ( mi->depth & MI_HASDEPTH_10 )
        factor = md->max_lut_size / 1024; 
    else
        factor = md->max_lut_size / 256; 
    /* factor = md->max_lut_size / (int) pow(2.0, (double) ms->depth); */
    steps = (double) (md->max_lut_size - 1);      /* depending on input size */
    mult = pow(2.0, (double) ms->depth) - 1.0;    /* depending on output */


    DBG(30, "calculate_gamma: factor=%d, mult =%f, steps=%f\n",
             factor, mult, steps);

    if ( strcmp(ms->val[OPT_GAMMA_MODE].s, MD_GAMMAMODE_SCALAR) == 0 )
      {
        exp = 1.0 / gamma;

        for ( i = 0; i < md->max_lut_size; i++ )  
          {
            val = (unsigned) (mult * pow((double) i / steps, exp) + .5);
            /* fprintf(stderr, "val[%4d]=%d\n", i, val); */ /* test */

            /* is the byte ordering correct ???? */
            if ( md->lut_entry_size == 2 ) 
                *((u_int16_t *) pos + i) = (u_int16_t) val;
            else
                *(pos + i) = (u_int8_t) val;
          }
      }
    else if ( strcmp(ms->val[OPT_GAMMA_MODE].s, MD_GAMMAMODE_LINEAR) == 0 )
      {
        for ( i = 0; i < md->max_lut_size; i++ )
          {
            if ( md->lut_entry_size == 2 )
                *((u_int16_t *) pos + i) = i / factor;
            else
                *(pos + i) = i / factor;
          }
      }
    return SANE_STATUS_GOOD;
}


/*---------- cancel_scan() ---------------------------------------------------*/

SANE_Status
cancel_scan(Microtek2_Scanner *ms)
{
    SANE_Status status;

    DBG(30, "cancel_scan: scanner=%p\n", ms);

    if ( ms->sfd != -1 ) 
      {
        ms->transfer_length = 0;
        /* wait_for_ready ?? */

        status = scsi_read_image(ms);
        sanei_scsi_close(ms->sfd);
        if ( status != SANE_STATUS_GOOD ) 
          {
            DBG(1, "cancel_scan: scsi_read_image status '%s'\n",
                     sane_strstatus(status));
            return status;
          }
      }

    status = SANE_STATUS_CANCELLED;

    if ( ms->src_buffer ) 
      {
        free((void *) ms->src_buffer);
        ms->src_buffer = NULL;
      }

    ms->sfd = -1;
    ms->scanning = SANE_FALSE;
    ms->cancelled = SANE_FALSE;
    ms->this_pass = 0; /* ???? */

    return status;
}


/*---------- eof_scan() ------------------------------------------------------*/

SANE_Status
eof_scan(Microtek2_Scanner *ms)
{

    DBG(30, "eof_scan: scanner=%p\n", ms);
    
    sanei_scsi_close(ms->sfd);
    if ( ms->src_buffer ) 
      {
        free((void *) ms->src_buffer);
        ms->src_buffer = NULL;
      }
    ms->sfd = -1;
    ms->scanning = SANE_FALSE;
    ms->cancelled = SANE_FALSE;
    ms->this_pass = 0; /* ???? */

    return SANE_STATUS_EOF;
}



/*---------- get_scan_mode_and_depth() ---------------------------------------*/

static SANE_Status
get_scan_mode_and_depth(Microtek2_Scanner *ms, int *mode, int *depth)
{
    /* This function translates the strings for the possible modes and */
    /* bitdepth into a more conveniant format as needed for SET WINDOW */
    
    DBG(30, "get_scan_mode_and_depth: handle=%p\n", ms);
    if ( strcmp(ms->val[OPT_MODE].s, MS_MODESTRING_COLOR36) == 0 )
      {
	*mode = MS_MODE_COLOR;
        *depth = 12;
      }
    else if ( strcmp(ms->val[OPT_MODE].s, MS_MODESTRING_COLOR30) == 0 ) 
      {
	*mode = MS_MODE_COLOR;
        *depth = 10;
      }
    else if ( strcmp(ms->val[OPT_MODE].s, MS_MODESTRING_COLOR24) == 0 ) 
      {
	*mode = MS_MODE_COLOR;
        *depth = 8;
      }
    else if ( strcmp(ms->val[OPT_MODE].s, MS_MODESTRING_GRAY12) == 0 ) 
      {
	*mode = MS_MODE_GRAY;
        *depth = 12;
      }
    else if ( strcmp(ms->val[OPT_MODE].s, MS_MODESTRING_GRAY10) == 0 ) 
      {
	*mode = MS_MODE_GRAY;
        *depth = 10;
      }
    else if ( strcmp(ms->val[OPT_MODE].s, MS_MODESTRING_GRAY8) == 0 ) 
      {
	*mode = MS_MODE_GRAY;
        *depth = 8;
      }
    else if ( strcmp(ms->val[OPT_MODE].s, MS_MODESTRING_HALFTONE) == 0) 
      {
	*mode = MS_MODE_HALFTONE;
        *depth = 1;
      }
    else if ( strcmp(ms->val[OPT_MODE].s, MS_MODESTRING_LINEART) == 0 )
      {
	*mode = MS_MODE_LINEART;
        *depth = 1;
      }
    else 
      {
        DBG(1, "get_scan_mode_and_depth: (B) Unknown mode %s\n",
                ms->val[OPT_MODE].s);
        return SANE_STATUS_INVAL;
      }
    DBG(30, "get_scan_mode_and_depth: mode=%d, depth=%d\n", *mode, *depth);
    return SANE_STATUS_GOOD;
}


/*---------- get_scan_parameters () ------------------------------------------*/

static SANE_Status
get_scan_parameters(Microtek2_Scanner *ms)
{
    Microtek2_Device *md;
    SANE_Fixed dpm = ms->dev->info.opt_resolution / MM_PER_INCH;
    SANE_Fixed x2_dots;
    SANE_Fixed y2_dots;
    int i;


    DBG(30, "get_scan_parameters: handle=%p\n", ms);

    md = ms->dev;
    get_scan_mode_and_depth(ms, &ms->mode, &ms->depth);
    
    if ( ms->mode == MS_MODE_HALFTONE )     /* select halftone pattern */
      {
        i = 0;
        while ( strcmp(md->halftone_mode_list[i], 
                       ms->val[OPT_HALFTONE_PATTERN].s) ) 
            ++i;
        ms->internal_ht_index = i;
      }

    if ( ms->mode == MS_MODE_LINEART )   
        ms->threshold = ms->val[OPT_THRESHOLD].w;

    DBG(30, "get_scan_parameters: mode=%d, depth=%d\n", ms->mode, ms->depth); 

    /* calculate positions, width and height in dots */
    ms->x1_dots = (SANE_Int) SANE_UNFIX(ms->val[OPT_TL_X].w * dpm + 0.5);
    ms->y1_dots = (SANE_Int) SANE_UNFIX(ms->val[OPT_TL_Y].w * dpm + 0.5);
    x2_dots = (SANE_Int) SANE_UNFIX(ms->val[OPT_BR_X].w * dpm + 0.5);
    y2_dots = (SANE_Int) SANE_UNFIX(ms->val[OPT_BR_Y].w * dpm + 0.5);
    ms->width_dots = abs(x2_dots - ms->x1_dots);
    ms->height_dots = abs(y2_dots - ms->y1_dots);
    /* ensure that the scan area is > 0 */
    if ( ms->height_dots == 0 )
        ms->height_dots = 10;
    if ( ms->width_dots == 0 )
        ms->width_dots = 10;

    ms->x_resolution_dpi = 
                    (SANE_Int) SANE_UNFIX(ms->val[OPT_RESOLUTION].w + 0.5);
    ms->y_resolution_dpi = 
                    (SANE_Int) SANE_UNFIX(ms->val[OPT_Y_RESOLUTION].w + 0.5);
    DBG(30, "get_scan_parameters: yres=%d, x1=%d, width=%d, y1=%d, height=%d\n",
             ms->y_resolution_dpi, ms->x1_dots, ms->width_dots, 
             ms->y1_dots, ms->height_dots);

    /* Preview mode */
    if ( ms->val[OPT_PREVIEW].w == SANE_TRUE )
      {
        ms->fastscan = SANE_TRUE;
        ms->quality = SANE_FALSE;
      }
    else 
      {
        ms->fastscan = SANE_FALSE;
        ms->quality = SANE_TRUE;
      }

    /* brightness, contrast, values 1,..,255 */
    ms->brightness_m = (SANE_Byte) (SANE_UNFIX(ms->val[OPT_BRIGHTNESS].w) 
                    / SANE_UNFIX(ms->percentage_range.max) * 254.0) + 1;
    /* ??? */
    ms->brightness_r = ms->brightness_g = ms->brightness_b = ms->brightness_m;

    ms->contrast_m = (SANE_Byte) (SANE_UNFIX(ms->val[OPT_CONTRAST].w) 
                    / SANE_UNFIX(ms->percentage_range.max) * 254.0) + 1;
    ms->contrast_r = ms->contrast_g = ms->contrast_b = ms->contrast_m;

    /* shadow, midtone, highlight */
    ms->shadow_m = ms->val[OPT_SHADOW].w;                
    ms->shadow_r = ms->val[OPT_SHADOW_R].w;                
    ms->shadow_g = ms->val[OPT_SHADOW_G].w;                
    ms->shadow_b = ms->val[OPT_SHADOW_B].w;
    ms->midtone_m = ms->val[OPT_MIDTONE].w;                
    ms->midtone_r = ms->val[OPT_MIDTONE_R].w;                
    ms->midtone_g = ms->val[OPT_MIDTONE_G].w;                
    ms->midtone_b = ms->val[OPT_MIDTONE_B].w;                
    ms->highlight_m = ms->val[OPT_HIGHLIGHT].w;     
    ms->highlight_r = ms->val[OPT_HIGHLIGHT_R].w;     
    ms->highlight_g = ms->val[OPT_HIGHLIGHT_G].w;     
    ms->highlight_b = ms->val[OPT_HIGHLIGHT_B].w;
  
    /* if ( scanner  supports d/l of gamma table ) */
    if ( ms->val[OPT_GAMMA_BIND].w == SANE_TRUE )
      {
        for ( i = 0; i < 3; i++ )
          {
            if ( ! md->info.onepass && ms->mode == MS_MODE_COLOR /*&& first pass*/)
                calculate_gamma(ms, ms->gamma_table[i] + SG_CMD_L,
                                SANE_UNFIX(ms->val[OPT_GAMMA_SCALAR].w));
            else 
                calculate_gamma(ms,
                                ms->gamma_table[0] + SG_CMD_L 
                                  + i * md->max_lut_size * md->lut_entry_size,
                                SANE_UNFIX(ms->val[OPT_GAMMA_SCALAR].w));
          }
      }
    else 
      {
        if ( ! md->info.onepass /*&& first pass*/)
          {
            calculate_gamma(ms, ms->gamma_table[0] + SG_CMD_L,
                            SANE_UNFIX(ms->val[OPT_GAMMA_SCALAR_R].w));
            calculate_gamma(ms, ms->gamma_table[1] + SG_CMD_L,
                            SANE_UNFIX(ms->val[OPT_GAMMA_SCALAR_G].w));
            calculate_gamma(ms, ms->gamma_table[2] + SG_CMD_L,
                            SANE_UNFIX(ms->val[OPT_GAMMA_SCALAR_B].w));
          }
        else
          {
            calculate_gamma(ms, ms->gamma_table[0] + SG_CMD_L, 
                            SANE_UNFIX(ms->val[OPT_GAMMA_SCALAR_R].w));
            calculate_gamma(ms,
                            ms->gamma_table[0] + SG_CMD_L 
                              + md->max_lut_size * md->lut_entry_size,
                            SANE_UNFIX(ms->val[OPT_GAMMA_SCALAR_G].w));
            calculate_gamma(ms,
                            ms->gamma_table[0] + SG_CMD_L 
                              + 2 * md->max_lut_size * md->lut_entry_size,
                            SANE_UNFIX(ms->val[OPT_GAMMA_SCALAR_B].w));
          }
      }

    return SANE_STATUS_GOOD;
}


/*---------- dump_area() -----------------------------------------------------*/

static SANE_Status
dump_area(u_int8_t *area, int len, char *info)
{
    /* this function dumps control or information blocks */
    
    #define BPL    16               /* bytes per line to print */
   
    int i;
    int o;
    int o_limit;

    if ( ! info[0] )
        info = "No additional info available";    

    DBG(30, "dump_area: %s\n", info);
   
    o_limit = (len + BPL - 1) / BPL;
    for ( o = 0; o < o_limit; o++) {
        fprintf(stderr, "  %4d: ", o * BPL); 
        for ( i=0; i < BPL && (o * BPL + i ) < len; i++) {
            if ( i == BPL / 2 ) 
                fprintf(stderr, " ");
            fprintf(stderr, "%02x", area[o * BPL + i]); 
        }
        
        fprintf(stderr, "%*s",  2 * ( 2 + BPL - i), " " );
        fprintf(stderr, "%s",  (i == BPL / 2) ? " " : "");

        for ( i = 0; i < BPL && (o * BPL + i ) < len; i++) {
            if ( i == BPL / 2 )
                fprintf(stderr, " ");
            fprintf(stderr, "%c", isprint(area[o * BPL + i])
                                  ? area[o * BPL + i]
                                  : '.');
        }
        fprintf(stderr, "\n");
    }

    return SANE_STATUS_GOOD;
}


/*---------- dump_area2() ----------------------------------------------------*/

static SANE_Status
dump_area2(u_int8_t *area, int len, char *info)
{
    int i;

    DBG(30, "dump_area2: %s\n", info);

    if ( ! info[0] )
        info = "No additional info available";    
    
    fprintf(stderr, "[%s]\n", info);

    for ( i = 0; i < len; i++)
        fprintf(stderr, "%02x", *(area + i));
    
    fprintf(stderr, "\n");

    return SANE_STATUS_GOOD;
}


/*---------- get_lut_size() --------------------------------------------------*/

static SANE_Status
get_lut_size(Microtek2_Info *mi, int *max_lut_size, int *lut_entry_size)
{
    /* returns the maximum lookup table size */

    DBG(30, "get_lut_size: mi=%p\n", mi);

    *max_lut_size = 0;
    *lut_entry_size = 0;

    if ( mi->lut_cap  &  MI_LUTCAP_256B )
      {
        *max_lut_size = 256;
        *lut_entry_size = 1;
      }
    if ( mi->lut_cap  &  MI_LUTCAP_1024B )
      {
        *max_lut_size = 1024;
        *lut_entry_size = 1;
      }
    if ( mi->lut_cap  &  MI_LUTCAP_1024W )
      {
        *max_lut_size = 1024;
        *lut_entry_size = 2;
      }
    if ( mi->lut_cap  &  MI_LUTCAP_4096B )
      {
          *max_lut_size = 4096;
          *lut_entry_size = 1;
      }
    if ( mi->lut_cap  &  MI_LUTCAP_4096W )
      {
          *max_lut_size = 4096;
          *lut_entry_size = 2;
      }
    DBG(30, "get_lut_size:  mi=%p, lut_size=%d, lut_word=%d\n",
             mi, *max_lut_size, *lut_entry_size);
    return SANE_STATUS_GOOD;
}


/*---------- init_options() --------------------------------------------------*/

static SANE_Status
init_options(Microtek2_Scanner *ms)
{
    SANE_Option_Descriptor *sod;
    Microtek2_Option_Value *val;
    Microtek2_Device *md;
    Microtek2_Info *mi;
    int i;


    DBG(30, "init_options: handle=%p\n", ms);

    sod = ms->sod;
    val = ms->val;
    md = ms->dev;
    mi = &md->info;


    sod[OPT_NUM_OPTS].name = SANE_NAME_NUM_OPTIONS;
    sod[OPT_NUM_OPTS].title = SANE_TITLE_NUM_OPTIONS;
    sod[OPT_NUM_OPTS].desc = SANE_DESC_NUM_OPTIONS;
    sod[OPT_NUM_OPTS].type = SANE_TYPE_INT;
    sod[OPT_NUM_OPTS].unit = SANE_UNIT_NONE;
    sod[OPT_NUM_OPTS].size = sizeof (SANE_Word);
    sod[OPT_NUM_OPTS].cap = SANE_CAP_SOFT_DETECT;
    sod[OPT_NUM_OPTS].constraint_type = SANE_CONSTRAINT_NONE;
    val[OPT_NUM_OPTS].w = NUM_OPTIONS;      /* NUM_OPTIONS is no option */

    /* The Scan Mode Group */
    /* sod[OPT_MODE_GROUP].name = ""; */
    sod[OPT_MODE_GROUP].title = "Scan Mode";
    /* sod[OPT_MODE_GROUP].desc = ""; */
    sod[OPT_MODE_GROUP].type = SANE_TYPE_GROUP;
    sod[OPT_MODE_GROUP].cap = 0;
    sod[OPT_MODE_GROUP].constraint_type = SANE_CONSTRAINT_NONE;

    i = 0;
    if ( (mi->scanmode & MI_HASMODE_COLOR) ) {
	if ( mi->depth & MI_HASDEPTH_12 )
	    ms->scanmode_list[i++] = MS_MODESTRING_COLOR36;
	if ( mi->depth & MI_HASDEPTH_10 )
	    ms->scanmode_list[i++] = MS_MODESTRING_COLOR30;
	ms->scanmode_list[i++] = MS_MODESTRING_COLOR24;
    }

    if ( mi->scanmode & MI_HASMODE_GRAY ) {
	if ( mi->depth & MI_HASDEPTH_12 )
	    ms->scanmode_list[i++] = MS_MODESTRING_GRAY12;
	if ( mi->depth & MI_HASDEPTH_10 )
	    ms->scanmode_list[i++] = MS_MODESTRING_GRAY10;
	ms->scanmode_list[i++] = MS_MODESTRING_GRAY8;
    }

    if ( mi->scanmode & MI_HASMODE_HALFTONE )
	ms->scanmode_list[i++] = MS_MODESTRING_HALFTONE;
    if ( mi->scanmode & MI_HASMODE_LINEART )
	ms->scanmode_list[i++] = MS_MODESTRING_LINEART;
    ms->scanmode_list[i] = NULL;

    sod[OPT_MODE].name = SANE_NAME_SCAN_MODE;
    sod[OPT_MODE].title = SANE_TITLE_SCAN_MODE;
    sod[OPT_MODE].desc = SANE_DESC_SCAN_MODE;
    sod[OPT_MODE].type = SANE_TYPE_STRING;
    sod[OPT_MODE].unit = SANE_UNIT_NONE;    /* ??? */
    sod[OPT_MODE].size = max_string_size(ms->scanmode_list);
    sod[OPT_MODE].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    sod[OPT_MODE].constraint_type = SANE_CONSTRAINT_STRING_LIST;
    sod[OPT_MODE].constraint.string_list = ms->scanmode_list;
    val[OPT_MODE].s = (SANE_String) strdup(ms->scanmode_list[0]);
   
    md->halftone_mode_list = malloc(13 * sizeof(SANE_String *));
    if ( md->halftone_mode_list == NULL ) 
      {
        DBG(1, "init_options: malloc for halftone_mode_list failed\n");
        return SANE_STATUS_NO_MEM;
      }
    md->halftone_mode_list[0] = "53-dot screen (53 gray levels)";
    md->halftone_mode_list[1] = "Horiz. screen (65 gray levels)";
    md->halftone_mode_list[2] = "Vert. screen (65 gray levels)";
    md->halftone_mode_list[3] = "Mixed page (33 gray levels)";
    md->halftone_mode_list[4] = "71-dot screen (29 gray levels)";
    md->halftone_mode_list[5] = "60-dot #1 (26 gray levels)";
    md->halftone_mode_list[6] = "60-dot #2 (26 gray levels)";
    md->halftone_mode_list[7] = "Fine detail #1 (17 gray levels)";
    md->halftone_mode_list[8] = "Fine detail #2 (17 gray levels)";
    md->halftone_mode_list[9] = "Slant line (17 gray levels)";
    md->halftone_mode_list[10] = "Posterizing (10 gray levels)";
    md->halftone_mode_list[11] = "High Contrast (5 gray levels)";
    md->halftone_mode_list[12] = NULL;
    sod[OPT_HALFTONE_PATTERN].name = SANE_NAME_HALFTONE;
    sod[OPT_HALFTONE_PATTERN].title = SANE_TITLE_HALFTONE;
    sod[OPT_HALFTONE_PATTERN].desc = SANE_DESC_HALFTONE_PATTERN;
    sod[OPT_HALFTONE_PATTERN].type = SANE_TYPE_STRING;
    sod[OPT_HALFTONE_PATTERN].unit = SANE_UNIT_NONE;    /* ??? */
    sod[OPT_HALFTONE_PATTERN].size = max_string_size(md->halftone_mode_list);
    sod[OPT_HALFTONE_PATTERN].cap  = SANE_CAP_SOFT_SELECT 
                                   | SANE_CAP_SOFT_DETECT
                                   | SANE_CAP_INACTIVE;
    sod[OPT_HALFTONE_PATTERN].constraint_type = SANE_CONSTRAINT_STRING_LIST;
    sod[OPT_HALFTONE_PATTERN].constraint.string_list = md->halftone_mode_list;
    val[OPT_HALFTONE_PATTERN].s = (SANE_String)
                                            strdup(md->halftone_mode_list[0]);

    md->threshold_range.min = 1;
    md->threshold_range.max = 255;
    md->threshold_range.quant = 1;
    sod[OPT_THRESHOLD].name = SANE_NAME_THRESHOLD;
    sod[OPT_THRESHOLD].title = SANE_TITLE_THRESHOLD;
    sod[OPT_THRESHOLD].desc = SANE_DESC_THRESHOLD;
    sod[OPT_THRESHOLD].type = SANE_TYPE_INT;
    sod[OPT_THRESHOLD].unit = SANE_UNIT_NONE;
    sod[OPT_THRESHOLD].size = sizeof(SANE_Int);
    sod[OPT_THRESHOLD].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT
                           | SANE_CAP_INACTIVE;
    sod[OPT_THRESHOLD].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_THRESHOLD].constraint.range = &md->threshold_range;
    val[OPT_THRESHOLD].w = 128;

    ms->x_res_range_dpi.min = SANE_FIX(10.0);
    ms->x_res_range_dpi.max = SANE_FIX(md->info.max_xresolution);
    ms->x_res_range_dpi.quant = SANE_FIX(1.0);
    sod[OPT_RESOLUTION].name = SANE_NAME_SCAN_RESOLUTION;
    sod[OPT_RESOLUTION].title = SANE_TITLE_SCAN_RESOLUTION;
    sod[OPT_RESOLUTION].desc = SANE_DESC_SCAN_RESOLUTION;
    sod[OPT_RESOLUTION].type = SANE_TYPE_FIXED;
    sod[OPT_RESOLUTION].unit = SANE_UNIT_DPI;
    sod[OPT_RESOLUTION].size = sizeof(SANE_Fixed);
    sod[OPT_RESOLUTION].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    sod[OPT_RESOLUTION].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_RESOLUTION].constraint.range = &ms->x_res_range_dpi;
    val[OPT_RESOLUTION].w = MIN(SANE_FIX(300.0), ms->x_res_range_dpi.max);
    
    ms->y_res_range_dpi.min = SANE_FIX(10.0);
    ms->y_res_range_dpi.max = SANE_FIX(md->info.max_yresolution);
    ms->y_res_range_dpi.quant = SANE_FIX(1.0);
    sod[OPT_Y_RESOLUTION].name = "y-" SANE_NAME_SCAN_RESOLUTION;
    sod[OPT_Y_RESOLUTION].title = "Y-Resolution";
    sod[OPT_Y_RESOLUTION].desc = SANE_DESC_SCAN_RESOLUTION;
    sod[OPT_Y_RESOLUTION].type = SANE_TYPE_FIXED;
    sod[OPT_Y_RESOLUTION].unit = SANE_UNIT_DPI;
    sod[OPT_Y_RESOLUTION].size = sizeof(SANE_Fixed);
    sod[OPT_Y_RESOLUTION].cap = SANE_CAP_SOFT_SELECT 
                                | SANE_CAP_SOFT_DETECT
                                | SANE_CAP_INACTIVE; 
    sod[OPT_Y_RESOLUTION].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_Y_RESOLUTION].constraint.range = &ms->y_res_range_dpi;
    val[OPT_Y_RESOLUTION].w = val[OPT_RESOLUTION].w; /* bind is default */

    sod[OPT_RESOLUTION_BIND].name = SANE_NAME_RESOLUTION_BIND;
    sod[OPT_RESOLUTION_BIND].title = SANE_TITLE_RESOLUTION_BIND;
    sod[OPT_RESOLUTION_BIND].desc = SANE_DESC_RESOLUTION_BIND;
    sod[OPT_RESOLUTION_BIND].type = SANE_TYPE_BOOL;
    sod[OPT_RESOLUTION_BIND].unit = SANE_UNIT_NONE;
    sod[OPT_RESOLUTION_BIND].size = sizeof(SANE_Bool);
    sod[OPT_RESOLUTION_BIND].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    sod[OPT_RESOLUTION_BIND].constraint_type = SANE_CONSTRAINT_NONE;
    val[OPT_RESOLUTION_BIND].w = SANE_TRUE;

    sod[OPT_PREVIEW].name = SANE_NAME_PREVIEW;
    sod[OPT_PREVIEW].title = SANE_TITLE_PREVIEW;
    sod[OPT_PREVIEW].desc = SANE_DESC_PREVIEW;
    sod[OPT_PREVIEW].type = SANE_TYPE_BOOL;
    sod[OPT_PREVIEW].unit = SANE_UNIT_NONE;
    sod[OPT_PREVIEW].size = sizeof(SANE_Bool);
    sod[OPT_PREVIEW].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    sod[OPT_PREVIEW].constraint_type = SANE_CONSTRAINT_NONE;
    val[OPT_PREVIEW].w = SANE_FALSE;

    /* Geometry group, for scan area selection */ 
    sod[OPT_GEOMETRY_GROUP].title = "Geometry";
    sod[OPT_GEOMETRY_GROUP].type = SANE_TYPE_GROUP;
    sod[OPT_GEOMETRY_GROUP].cap = SANE_CAP_ADVANCED;
    sod[OPT_GEOMETRY_GROUP].constraint_type = SANE_CONSTRAINT_NONE;

    ms->x_range_mm.min = SANE_FIX(0.0);
    ms->x_range_mm.max = SANE_FIX((double) md->info.geo_width 
                                  / (double) md->info.opt_resolution
                                  * MM_PER_INCH);
    ms->x_range_mm.quant = SANE_FIX(0.0);
    
    ms->y_range_mm.min = SANE_FIX(0.0);
    ms->y_range_mm.max = SANE_FIX((double) md->info.geo_height 
                                  / (double) md->info.opt_resolution
                                  * MM_PER_INCH);
    ms->y_range_mm.quant = SANE_FIX(0.0);
    sod[OPT_TL_X].name = SANE_NAME_SCAN_TL_X;
    sod[OPT_TL_X].title = SANE_TITLE_SCAN_TL_X;
    sod[OPT_TL_X].desc = SANE_DESC_SCAN_TL_X;
    sod[OPT_TL_X].type = SANE_TYPE_FIXED;
    sod[OPT_TL_X].unit = SANE_UNIT_MM;
    sod[OPT_TL_X].size = sizeof(SANE_Fixed);
    sod[OPT_TL_X].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    sod[OPT_TL_X].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_TL_X].constraint.range = &ms->x_range_mm;
    val[OPT_TL_X].w = SANE_FIX(0.0);

    sod[OPT_TL_Y].name = SANE_NAME_SCAN_TL_Y;
    sod[OPT_TL_Y].title = SANE_TITLE_SCAN_TL_Y;
    sod[OPT_TL_Y].desc = SANE_DESC_SCAN_TL_Y;
    sod[OPT_TL_Y].type = SANE_TYPE_FIXED;
    sod[OPT_TL_Y].unit = SANE_UNIT_MM;
    sod[OPT_TL_Y].size = sizeof(SANE_Fixed);
    sod[OPT_TL_Y].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    sod[OPT_TL_Y].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_TL_Y].constraint.range = &ms->y_range_mm;
    val[OPT_TL_Y].w = SANE_FIX(0.0);

    sod[OPT_BR_X].name = SANE_NAME_SCAN_BR_X;
    sod[OPT_BR_X].title = SANE_TITLE_SCAN_BR_X;
    sod[OPT_BR_X].desc = SANE_DESC_SCAN_BR_X;
    sod[OPT_BR_X].type = SANE_TYPE_FIXED;
    sod[OPT_BR_X].unit = SANE_UNIT_MM;
    sod[OPT_BR_X].size = sizeof(SANE_Fixed);
    sod[OPT_BR_X].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    sod[OPT_BR_X].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_BR_X].constraint.range = &ms->x_range_mm;
    val[OPT_BR_X].w = ms->x_range_mm.max;
    
    sod[OPT_BR_Y].name = SANE_NAME_SCAN_BR_Y;
    sod[OPT_BR_Y].title = SANE_TITLE_SCAN_BR_Y;
    sod[OPT_BR_Y].desc = SANE_DESC_SCAN_BR_Y;
    sod[OPT_BR_Y].type = SANE_TYPE_FIXED;
    sod[OPT_BR_Y].unit = SANE_UNIT_MM;
    sod[OPT_BR_Y].size = sizeof(SANE_Fixed);
    sod[OPT_BR_Y].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    sod[OPT_BR_Y].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_BR_Y].constraint.range = &ms->y_range_mm;
    val[OPT_BR_Y].w = ms->y_range_mm.max;

    ms->percentage_range.min = 0 << SANE_FIXED_SCALE_SHIFT;
    ms->percentage_range.max = 200 << SANE_FIXED_SCALE_SHIFT;
    ms->percentage_range.quant = 1 << SANE_FIXED_SCALE_SHIFT;
    sod[OPT_ENHANCEMENT_GROUP].title = "Enhancement";
    sod[OPT_ENHANCEMENT_GROUP].desc = "";
    sod[OPT_ENHANCEMENT_GROUP].type = SANE_TYPE_GROUP;
    sod[OPT_ENHANCEMENT_GROUP].cap = 0;
    sod[OPT_ENHANCEMENT_GROUP].constraint_type = SANE_CONSTRAINT_NONE;

    sod[OPT_BRIGHTNESS].name = SANE_NAME_BRIGHTNESS;
    sod[OPT_BRIGHTNESS].title = SANE_TITLE_BRIGHTNESS;
    sod[OPT_BRIGHTNESS].desc = SANE_DESC_BRIGHTNESS;
    sod[OPT_BRIGHTNESS].type = SANE_TYPE_FIXED;
    sod[OPT_BRIGHTNESS].unit = SANE_UNIT_PERCENT;
    sod[OPT_BRIGHTNESS].size = sizeof(SANE_Fixed);
    sod[OPT_BRIGHTNESS].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    sod[OPT_BRIGHTNESS].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_BRIGHTNESS].constraint.range = &ms->percentage_range;
    val[OPT_BRIGHTNESS].w = 100 << SANE_FIXED_SCALE_SHIFT;

    sod[OPT_CONTRAST].name = SANE_NAME_CONTRAST;
    sod[OPT_CONTRAST].title = SANE_TITLE_CONTRAST;
    sod[OPT_CONTRAST].desc = SANE_DESC_CONTRAST;
    sod[OPT_CONTRAST].type = SANE_TYPE_FIXED;
    sod[OPT_CONTRAST].unit = SANE_UNIT_PERCENT;
    sod[OPT_CONTRAST].size = sizeof(SANE_Fixed);
    sod[OPT_CONTRAST].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    sod[OPT_CONTRAST].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_CONTRAST].constraint.range = &ms->percentage_range;
    val[OPT_CONTRAST].w = 100 << SANE_FIXED_SCALE_SHIFT;

    sod[OPT_GAMMA_GROUP].title = "Gamma";
    sod[OPT_GAMMA_GROUP].desc = "";
    sod[OPT_GAMMA_GROUP].type = SANE_TYPE_GROUP;
    sod[OPT_GAMMA_GROUP].cap = 0;
    sod[OPT_GAMMA_GROUP].constraint_type = SANE_CONSTRAINT_NONE;

    md->gammamode_list[0] = MD_GAMMAMODE_LINEAR;   /* must come first */
    md->gammamode_list[1] = MD_GAMMAMODE_SCALAR;
    /* md->gammamode_list[2] = MD_GAMMAMODE_CUSTOM; */
    /* md->gammamode_list[3] = NULL; */
    md->gammamode_list[2] = NULL;
    sod[OPT_GAMMA_MODE].name = "gammamode";
    sod[OPT_GAMMA_MODE].title = "Gamma mode";
    sod[OPT_GAMMA_MODE].desc = "";
    sod[OPT_GAMMA_MODE].type = SANE_TYPE_STRING;
    sod[OPT_GAMMA_MODE].unit = SANE_UNIT_NONE;    /* ??? */
    sod[OPT_GAMMA_MODE].size = max_string_size(md->gammamode_list);
    sod[OPT_GAMMA_MODE].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    /* if the scanner doesn't support d/l of lookup tables deactivate */
    /* this option */
    if ( md->max_lut_size == 0 )
        sod[OPT_GAMMA_MODE].cap |= SANE_CAP_INACTIVE;
    sod[OPT_GAMMA_MODE].constraint_type = SANE_CONSTRAINT_STRING_LIST;
    sod[OPT_GAMMA_MODE].constraint.string_list = md->gammamode_list;
    val[OPT_GAMMA_MODE].s = (SANE_String) strdup(md->gammamode_list[0]);

    sod[OPT_GAMMA_BIND].name = "gamma-bind";
    sod[OPT_GAMMA_BIND].title = "Bind gamma";
    sod[OPT_GAMMA_BIND].desc = "";
    sod[OPT_GAMMA_BIND].type = SANE_TYPE_BOOL;
    sod[OPT_GAMMA_BIND].unit = SANE_UNIT_NONE;
    sod[OPT_GAMMA_BIND].size = sizeof(SANE_Bool);
    sod[OPT_GAMMA_BIND].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT
                            | SANE_CAP_INACTIVE;
    sod[OPT_GAMMA_BIND].constraint_type = SANE_CONSTRAINT_NONE;
    val[OPT_GAMMA_BIND].w = SANE_TRUE;

    md->gamma_range.min = SANE_FIX(0.1);
    md->gamma_range.max = SANE_FIX(4.0);
    md->gamma_range.quant = SANE_FIX(0.1);

    /* this is active if gamma_bind == true and gammamode == scalar */
    sod[OPT_GAMMA_SCALAR].name = "scalar-gamma";
    sod[OPT_GAMMA_SCALAR].title = "Scalar gamma";
    sod[OPT_GAMMA_SCALAR].desc = "";
    sod[OPT_GAMMA_SCALAR].type = SANE_TYPE_FIXED;
    sod[OPT_GAMMA_SCALAR].unit = SANE_UNIT_NONE;
    sod[OPT_GAMMA_SCALAR].size = sizeof(SANE_Fixed);
    sod[OPT_GAMMA_SCALAR].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT
                              | SANE_CAP_INACTIVE;
    sod[OPT_GAMMA_SCALAR].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_GAMMA_SCALAR].constraint.range = &md->gamma_range;
    val[OPT_GAMMA_SCALAR].w = SANE_FIX(1.8);

    sod[OPT_GAMMA_SCALAR_R].name = "red-scalar-gamma";
    sod[OPT_GAMMA_SCALAR_R].title = "Scalar gamma red";
    sod[OPT_GAMMA_SCALAR_R].desc = "";
    sod[OPT_GAMMA_SCALAR_R].type = SANE_TYPE_FIXED;
    sod[OPT_GAMMA_SCALAR_R].unit = SANE_UNIT_NONE;
    sod[OPT_GAMMA_SCALAR_R].size = sizeof(SANE_Fixed);
    sod[OPT_GAMMA_SCALAR_R].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT
                              | SANE_CAP_INACTIVE;
    sod[OPT_GAMMA_SCALAR_R].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_GAMMA_SCALAR_R].constraint.range = &md->gamma_range;
    val[OPT_GAMMA_SCALAR_R].w = SANE_FIX(1.8);

    sod[OPT_GAMMA_SCALAR_G].name = "green-scalar-gamma";
    sod[OPT_GAMMA_SCALAR_G].title = "Scalar gamma green";
    sod[OPT_GAMMA_SCALAR_G].desc = "";
    sod[OPT_GAMMA_SCALAR_G].type = SANE_TYPE_FIXED;
    sod[OPT_GAMMA_SCALAR_G].unit = SANE_UNIT_NONE;
    sod[OPT_GAMMA_SCALAR_G].size = sizeof(SANE_Fixed);
    sod[OPT_GAMMA_SCALAR_G].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT
                              | SANE_CAP_INACTIVE;
    sod[OPT_GAMMA_SCALAR_G].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_GAMMA_SCALAR_G].constraint.range = &md->gamma_range;
    val[OPT_GAMMA_SCALAR_G].w = SANE_FIX(1.8);

    sod[OPT_GAMMA_SCALAR_B].name = "blue-scalar-gamma";
    sod[OPT_GAMMA_SCALAR_B].title = "Scalar gamma blue";
    sod[OPT_GAMMA_SCALAR_B].desc = "";
    sod[OPT_GAMMA_SCALAR_B].type = SANE_TYPE_FIXED;
    sod[OPT_GAMMA_SCALAR_B].unit = SANE_UNIT_NONE;
    sod[OPT_GAMMA_SCALAR_B].size = sizeof(SANE_Fixed);
    sod[OPT_GAMMA_SCALAR_B].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT
                              | SANE_CAP_INACTIVE;
    sod[OPT_GAMMA_SCALAR_B].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_GAMMA_SCALAR_B].constraint.range = &md->gamma_range;
    val[OPT_GAMMA_SCALAR_B].w = SANE_FIX(1.8);

    sod[OPT_SMH_GROUP].title = "Shadow, midtone, highlight";
    sod[OPT_SMH_GROUP].desc = "";
    sod[OPT_SMH_GROUP].type = SANE_TYPE_GROUP;
    sod[OPT_SMH_GROUP].cap = 0;
    sod[OPT_SMH_GROUP].constraint_type = SANE_CONSTRAINT_NONE;

    md->channel_list[0] = MD_CHANNEL_MASTER;
    md->channel_list[1] = MD_CHANNEL_RED;
    md->channel_list[2] = MD_CHANNEL_GREEN;
    md->channel_list[3] = MD_CHANNEL_BLUE;
    md->channel_list[4] = NULL;
    sod[OPT_CHANNEL].name = "channel";
    sod[OPT_CHANNEL].title = "Channel";
    sod[OPT_CHANNEL].desc = "";
    sod[OPT_CHANNEL].type = SANE_TYPE_STRING;
    sod[OPT_CHANNEL].unit = SANE_UNIT_NONE;    /* ??? */
    sod[OPT_CHANNEL].size = max_string_size(md->channel_list);
    sod[OPT_CHANNEL].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    sod[OPT_CHANNEL].constraint_type = SANE_CONSTRAINT_STRING_LIST;
    sod[OPT_CHANNEL].constraint.string_list = md->channel_list;
    val[OPT_CHANNEL].s = strdup((SANE_String) md->channel_list[0]);

    md->shadow_range.min = 0;
    md->shadow_range.max = 253;
    md->shadow_range.quant = 1;
    sod[OPT_SHADOW].name = "shadow";
    sod[OPT_SHADOW].title = "Shadow";
    sod[OPT_SHADOW].desc = "";
    sod[OPT_SHADOW].type = SANE_TYPE_INT;
    sod[OPT_SHADOW].unit = SANE_UNIT_NONE;
    sod[OPT_SHADOW].size = sizeof(SANE_Int);
    sod[OPT_SHADOW].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    sod[OPT_SHADOW].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_SHADOW].constraint.range = &md->shadow_range;
    val[OPT_SHADOW].w = 0;

    sod[OPT_SHADOW_R].name = "red-shadow";
    sod[OPT_SHADOW_R].title = "Shadow red";
    sod[OPT_SHADOW_R].desc = "";
    sod[OPT_SHADOW_R].type = SANE_TYPE_INT;
    sod[OPT_SHADOW_R].unit = SANE_UNIT_NONE;
    sod[OPT_SHADOW_R].size = sizeof(SANE_Int);
    sod[OPT_SHADOW_R].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT
                          | SANE_CAP_INACTIVE;
    sod[OPT_SHADOW_R].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_SHADOW_R].constraint.range = &md->shadow_range;
    val[OPT_SHADOW_R].w = 0;

    sod[OPT_SHADOW_G].name = "green-shadow";
    sod[OPT_SHADOW_G].title = "Shadow green";
    sod[OPT_SHADOW_G].desc = "";
    sod[OPT_SHADOW_G].type = SANE_TYPE_INT;
    sod[OPT_SHADOW_G].unit = SANE_UNIT_NONE;
    sod[OPT_SHADOW_G].size = sizeof(SANE_Int);
    sod[OPT_SHADOW_G].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT
                          | SANE_CAP_INACTIVE;
    sod[OPT_SHADOW_G].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_SHADOW_G].constraint.range = &md->shadow_range;
    val[OPT_SHADOW_G].w = 0;

    sod[OPT_SHADOW_B].name = "blue-shadow";
    sod[OPT_SHADOW_B].title = "Shadow blue";
    sod[OPT_SHADOW_B].desc = "";
    sod[OPT_SHADOW_B].type = SANE_TYPE_INT;
    sod[OPT_SHADOW_B].unit = SANE_UNIT_NONE;
    sod[OPT_SHADOW_B].size = sizeof(SANE_Int);
    sod[OPT_SHADOW_B].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT
                          | SANE_CAP_INACTIVE;
    sod[OPT_SHADOW_B].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_SHADOW_B].constraint.range = &md->shadow_range;
    val[OPT_SHADOW_B].w = 0;

    md->midtone_range.min = 1;
    md->midtone_range.max = 254;
    md->midtone_range.quant = 1;
    sod[OPT_MIDTONE].name = "midtone";
    sod[OPT_MIDTONE].title = "Midtone";
    sod[OPT_MIDTONE].desc = "";
    sod[OPT_MIDTONE].type = SANE_TYPE_INT;
    sod[OPT_MIDTONE].unit = SANE_UNIT_NONE;
    sod[OPT_MIDTONE].size = sizeof(SANE_Int);
    sod[OPT_MIDTONE].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    sod[OPT_MIDTONE].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_MIDTONE].constraint.range = &md->midtone_range;
    val[OPT_MIDTONE].w = 128;

    sod[OPT_MIDTONE_R].name = "red-midtone";
    sod[OPT_MIDTONE_R].title = "Midtone red";
    sod[OPT_MIDTONE_R].desc = "";
    sod[OPT_MIDTONE_R].type = SANE_TYPE_INT;
    sod[OPT_MIDTONE_R].unit = SANE_UNIT_NONE;
    sod[OPT_MIDTONE_R].size = sizeof(SANE_Int);
    sod[OPT_MIDTONE_R].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT
                           | SANE_CAP_INACTIVE;
    sod[OPT_MIDTONE_R].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_MIDTONE_R].constraint.range = &md->midtone_range;
    val[OPT_MIDTONE_R].w = 128;

    sod[OPT_MIDTONE_G].name = "green-midtone";
    sod[OPT_MIDTONE_G].title = "Midtone green";
    sod[OPT_MIDTONE_G].desc = "";
    sod[OPT_MIDTONE_G].type = SANE_TYPE_INT;
    sod[OPT_MIDTONE_G].unit = SANE_UNIT_NONE;
    sod[OPT_MIDTONE_G].size = sizeof(SANE_Int);
    sod[OPT_MIDTONE_G].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT
                           | SANE_CAP_INACTIVE;
    sod[OPT_MIDTONE_G].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_MIDTONE_G].constraint.range = &md->midtone_range;
    val[OPT_MIDTONE_G].w = 128;

    sod[OPT_MIDTONE_B].name = "blue-midtone";
    sod[OPT_MIDTONE_B].title = "Midtone blue";
    sod[OPT_MIDTONE_B].desc = "";
    sod[OPT_MIDTONE_B].type = SANE_TYPE_INT;
    sod[OPT_MIDTONE_B].unit = SANE_UNIT_NONE;
    sod[OPT_MIDTONE_B].size = sizeof(SANE_Int);
    sod[OPT_MIDTONE_B].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT
                           | SANE_CAP_INACTIVE;
    sod[OPT_MIDTONE_B].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_MIDTONE_B].constraint.range = &md->midtone_range;
    val[OPT_MIDTONE_B].w = 128;

    md->highlight_range.min = 2;
    md->highlight_range.max = 255;
    md->highlight_range.quant = 1;
    sod[OPT_HIGHLIGHT].name = "highlight";
    sod[OPT_HIGHLIGHT].title = "Highlight";
    sod[OPT_HIGHLIGHT].desc = "";
    sod[OPT_HIGHLIGHT].type = SANE_TYPE_INT;
    sod[OPT_HIGHLIGHT].unit = SANE_UNIT_NONE;
    sod[OPT_HIGHLIGHT].size = sizeof(SANE_Int);
    sod[OPT_HIGHLIGHT].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
    sod[OPT_HIGHLIGHT].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_HIGHLIGHT].constraint.range = &md->highlight_range;
    val[OPT_HIGHLIGHT].w = 255;

    sod[OPT_HIGHLIGHT_R].name = "red-highlight";
    sod[OPT_HIGHLIGHT_R].title = "Highlight red";
    sod[OPT_HIGHLIGHT_R].desc = "";
    sod[OPT_HIGHLIGHT_R].type = SANE_TYPE_INT;
    sod[OPT_HIGHLIGHT_R].unit = SANE_UNIT_NONE;
    sod[OPT_HIGHLIGHT_R].size = sizeof(SANE_Int);
    sod[OPT_HIGHLIGHT_R].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT
                             | SANE_CAP_INACTIVE;
    sod[OPT_HIGHLIGHT_R].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_HIGHLIGHT_R].constraint.range = &md->highlight_range;
    val[OPT_HIGHLIGHT_R].w = 255;

    sod[OPT_HIGHLIGHT_G].name = "green-highlight";
    sod[OPT_HIGHLIGHT_G].title = "Highlight green";
    sod[OPT_HIGHLIGHT_G].desc = "";
    sod[OPT_HIGHLIGHT_G].type = SANE_TYPE_INT;
    sod[OPT_HIGHLIGHT_G].unit = SANE_UNIT_NONE;
    sod[OPT_HIGHLIGHT_G].size = sizeof(SANE_Int);
    sod[OPT_HIGHLIGHT_G].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT
                             | SANE_CAP_INACTIVE;
    sod[OPT_HIGHLIGHT_G].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_HIGHLIGHT_G].constraint.range = &md->highlight_range;
    val[OPT_HIGHLIGHT_G].w = 255;

    sod[OPT_HIGHLIGHT_B].name = "blue-highlight";
    sod[OPT_HIGHLIGHT_B].title = "Highlight blue";
    sod[OPT_HIGHLIGHT_B].desc = "";
    sod[OPT_HIGHLIGHT_B].type = SANE_TYPE_INT;
    sod[OPT_HIGHLIGHT_B].unit = SANE_UNIT_NONE;
    sod[OPT_HIGHLIGHT_B].size = sizeof(SANE_Int);
    sod[OPT_HIGHLIGHT_B].cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT
                             | SANE_CAP_INACTIVE;
    sod[OPT_HIGHLIGHT_B].constraint_type = SANE_CONSTRAINT_RANGE;
    sod[OPT_HIGHLIGHT_B].constraint.range = &md->highlight_range;
    val[OPT_HIGHLIGHT_B].w = 255;

    return SANE_STATUS_GOOD;
}


/*---------- max_string_size() -----------------------------------------------*/

static size_t
max_string_size (const SANE_String_Const strings[])
{
  size_t size;
  size_t max_size = 0;
  int i;

  for (i = 0; strings[i]; ++i) {
    size = strlen(strings[i]) + 1; /* +1 because NUL counts as part of string */
    if (size > max_size) max_size = size;
  }
  return max_size;
}


/*---------- proc_gray_data() ------------------------------------------------*/

static SANE_Status                
proc_gray_data(Microtek2_Scanner *ms, SANE_Byte *buf, SANE_Int *len)
{
    SANE_Int lines;
    SANE_Byte *src_buf;
    SANE_Byte *dest_buf;
    int line;
    int pixel;
    int pad;
    int bpl;                    /* bytes per line */
    int bpp;                    /* bytes per pixel */
    int ppl;                    /* pixel per line */


    DBG(30, "proc_gray_data: ms=%p, buf=%p\n", ms, buf);
             
    lines = ms->src_lines_to_read;
    bpl = ms->bpl;
    ppl = ms->ppl;

    /* if the pixels per line are an odd number there may be a padding byte */
    /* at the end of the line, because transfers are always word aligned */
    pad = ppl % 2;
    
    bpp = ( bpl - pad ) / ppl;
    src_buf = ms->src_buffer;
    dest_buf = buf;
     
    line = 0;
    do {
        for ( pixel = 0; pixel < ppl; pixel++ ) {     
            memcpy(dest_buf, src_buf + pixel * bpp, bpp);
            dest_buf += bpp;
        }

        src_buf += bpl;
        ++line;
    } while ( line < lines );

    *len = (bpl - pad) * lines;

    DBG(30, "proc_gray_data: lines=%d, bytes_per_line=%d, len=%d\n",
             lines, bpl, *len); 
    return SANE_STATUS_GOOD;
}



/*---------- proc_onebit_data() ----------------------------------------------*/

static SANE_Status                
proc_onebit_data(Microtek2_Scanner *ms, SANE_Byte *buf, SANE_Int *len)
{
    SANE_Byte *src;
    SANE_Byte *dest;
    int bytes_to_copy;   /* bytes per line to copy into the frontend buffer */
    int line;


    DBG(30, "proc_onebit_data: ms=%p, buf=%p\n", ms, buf);

    /* the scanner alwas seems to transfer an even number of bytes */
    /* use ppl to calculate the number of bytes */
    bytes_to_copy = (ms->ppl + 7) / 8; 
    src = ms->src_buffer;
    dest = buf;

    for ( line = 0; line < ms->src_lines_to_read; line++ )
      {
        memcpy(dest, src, bytes_to_copy);
        dest += bytes_to_copy;
        src += ms->bpl;
      }
 
    *len = bytes_to_copy * ms->src_lines_to_read;

    return SANE_STATUS_GOOD;

}


/*---------- proc_segreg_data() ----------------------------------------------*/

static SANE_Status                
proc_segreg_data(Microtek2_Scanner *ms, 
                 SANE_Byte *buf,
                 SANE_Int maxlen,
                 SANE_Int *len)
{
    Microtek2_Info *mi;
    SANE_Byte *src_buf;
    SANE_Byte *prehold_buf[3];
    int bpp;                    /* bytes per pixel */
    int bpf;                    /* bytes per frame including color indicator */
    int pad;
    int pixel;
    int color_seq2;
    int color;
    int srcframes[3];
    int prehold[3];             /* frames in prehold buffer */
    int from_prehold[3];        /* that many frames must be copied from */
    int from_src[3];            /* from the prehold buffer and the src buffer */
    int lines_to_deliver;
    int dest_fit_lines;
    int lines;
    int frame;


    DBG(30, "proc_segreg_data: bpl=%d, lines=%d\n",
             ms->bpl, ms->src_lines_to_read);

    /* given: number of red, green, blue planes in the prehold buffer, */
    /* ms->prehold[i], and a pointer to the prehold buffer, ms->prehold */


    mi = &ms->dev->info;
    src_buf = ms->src_buffer;
    lines = ms->src_lines_to_read;

    pad = ((ms->bpl - 2 * 3) / 3 ) % ms->ppl;   /* ?????? */
    bpp = (ms->bpl - 3 * (2 + pad)) / ms->ppl / 3;
    bpf = ms->bpl / 3;

    for ( color = 0; color < 3; color++ )
      {
        srcframes[color] = 0;
        prehold[color] = ms->prehold[color];
        prehold_buf[color] = ms->prehold_buf
                             + color * (ms->size_of_prehold_buf / 3);
        DBG(30, "proc_segreg_data: prehold_buf[%d]=%p\n", 
                 color, prehold_buf[color]);
      }

    color_seq2 = mi->color_sequence[2];   
    for ( frame = 0; frame < lines * 3; frame++ )
      {
        switch ( *(src_buf + frame *  bpf) )
          {
            case 'R':
              ++srcframes[MI_COLOR_RED]; break;
            case 'G':
              ++srcframes[MI_COLOR_GREEN]; break;
            case 'B':
              ++srcframes[MI_COLOR_BLUE]; break;
            default:
              DBG(1, "proc_segreg_data: unknown color indicator (1) 0x%02x\n",
                      *(src_buf + frame * bpf));
              break;
          }
      }
   
    dest_fit_lines = maxlen / (ms->bpl - 3 * ( 2 + pad ));
    lines_to_deliver = MIN(srcframes[color_seq2] + prehold[color_seq2], 
                           dest_fit_lines);  /* lines_to_deliver maybe zero */

    DBG(30, "proc_sgreg_data: fitlines=%d, to_deliver=%d bpp=%d bpf=%d
             srcframes0=%d, srcframes1=%d, srcframes2=%d
             prehold0=%d, prehold1=%d, prehold2=%d\n", dest_fit_lines, 
             lines_to_deliver, bpp, bpf, srcframes[0], srcframes[1], 
             srcframes[2], prehold[0], prehold[1], prehold[2]); 
            
    if ( dest_fit_lines == 0 ) 
      {
        DBG(1,"proc_segreg_data: dest_fit_lines == 0, maxlen %d to small\n",
               maxlen);
        return SANE_STATUS_INVAL;   /* ???? */
      }
    
    for ( color = 0; color < 3; color++ )
      {
        SANE_Byte *from;
        SANE_Byte *dest;

        dest = buf;
        from_prehold[color] = MIN(prehold[color], lines_to_deliver);
        from_src[color] = lines_to_deliver - from_prehold[color];    
        DBG(30, "proc_segregdata: from[prehold[%d]=%d, from_src=%d\n",
                 color, from_prehold[color], from_src[color]);
        
        for ( frame = 0; frame < from_prehold[color]; frame++ )
          {
            from = prehold_buf[color] + frame * bpf + 2;
            for ( pixel = 0; pixel < ms->ppl; pixel++ )
              {
                memcpy(dest + bpp * color, from, bpp);
                from += bpp;
                dest += 3 * bpp;
              }
          }    

        /* copy the not copied frames to the beginning of prehold_buf[color] */
        if ( prehold[color] - from_prehold[color] > 0 )
            memmove(prehold_buf[color], 
                    prehold_buf[color] + from_prehold[color] * bpf,
                    (prehold[color] - from_prehold[color]) * bpf);

        prehold[color] -= from_prehold[color];


        /* now copy from_source[color] frames from the source buffer */

        from = src_buf;
        for ( frame = 0; frame < from_src[color]; frame++ )
          {
            int found = 0;

            /* search next frame of current color, this code could */
            /* be shorter */
            for ( ; from < src_buf + lines * ms->bpl; from += bpf ) 
              { 
                switch ( *from )
                  {
                    case 'R': 
                      if ( color == MI_COLOR_RED ) { found = 1; }
                      break;
                    case 'G': 
                      if ( color == MI_COLOR_GREEN ) { found = 1; }
                      break;
                    case 'B': 
                      if ( color == MI_COLOR_BLUE ) { found = 1; }
                      break;
                    default:
                      DBG(1, "proc_segreg_data: unknown color indicator (2)"
                             " 0x%02x\n", *from);
                      break;
                  }
                if ( found ) 
                    break;
              }

            if ( ! found )
              {
                DBG(1, "proc_segreg_data: color indicator not found\n");
                return SANE_STATUS_INVAL;
              }
            
            /* from += 2; */   /* skip color indicator */
            for ( pixel = 0; pixel < ms->ppl; pixel++ )
              {
                memcpy(dest + bpp * color, from + pixel * bpp + 2, bpp);
                dest += 3 * bpp;
              }
            from += bpf;
          }  

        /* copy the rest of the frames of this color from the source */
        /* buffer into the prehold buffer */

        /* from points to the last processed frame of this color */

        /* from += bpf; */
        dest = prehold_buf[color] + prehold[color] * bpf;

        /* could be shorter */
        for ( frame = from_src[color]; frame < srcframes[color];  )
          {
            switch (*from)
              {
                case 'R':
                  if ( color == MI_COLOR_RED ) 
                    {
                      memcpy(dest, from, bpf);
                      dest += bpf;
                      ++frame;
                    }
                  break;
                case 'G':
                  if ( color == MI_COLOR_GREEN )
                    {
                      memcpy(dest, from, bpf);
                      dest += bpf;
                      ++frame;
                    }
                  break;
                case 'B':
                  if ( color == MI_COLOR_BLUE )
                    {
                      memcpy(dest, from, bpf);
                      dest += bpf;
                      ++frame;
                    }
                  break;
                default:
                  DBG(1, "proc_segreg_data: unknown color indicator (3) 
                          0x%02x\n",  *from);
                  break;
              }
            from += bpf;
          }
          DBG(30,"proc_segreg_data: srcframes[%d]=%d, from_src=%d\n", 
                  color, srcframes[color], from_src[color]); 
              
        ms->prehold[color] = 
                  prehold[color] + srcframes[color] - from_src[color];
      } 
    ms->undelivered_lines = MAX(ms->prehold[0], 
                                MAX(ms->prehold[1], ms->prehold[2]));
    *len = lines_to_deliver * ( ms->bpl - 3 * ( 2 + pad ));
    return SANE_STATUS_GOOD;
}


/*---------- restore_gamma_options() -----------------------------------------*/

static SANE_Status                
restore_gamma_options(SANE_Option_Descriptor *sod, 
                      Microtek2_Option_Value *val, 
                      int mode)
{
    switch( mode )
      {
        case MS_MODE_COLOR:
          sod[OPT_GAMMA_MODE].cap &= ~SANE_CAP_INACTIVE;
          if ( strcmp(val[OPT_GAMMA_MODE].s, MD_GAMMAMODE_LINEAR) == 0 )
            {
              sod[OPT_GAMMA_BIND].cap |= SANE_CAP_INACTIVE;
              sod[OPT_GAMMA_SCALAR].cap |= SANE_CAP_INACTIVE;
              sod[OPT_GAMMA_SCALAR_R].cap |= SANE_CAP_INACTIVE;
              sod[OPT_GAMMA_SCALAR_G].cap |= SANE_CAP_INACTIVE;
              sod[OPT_GAMMA_SCALAR_B].cap |= SANE_CAP_INACTIVE;
            }
          else if ( strcmp(val[OPT_GAMMA_MODE].s, MD_GAMMAMODE_SCALAR) == 0 )
            {
              sod[OPT_GAMMA_BIND].cap &= ~SANE_CAP_INACTIVE;
              if ( val[OPT_GAMMA_BIND].w == SANE_TRUE ) 
                {
                  sod[OPT_GAMMA_SCALAR].cap &= ~SANE_CAP_INACTIVE;
                  sod[OPT_GAMMA_SCALAR_R].cap |= SANE_CAP_INACTIVE;
                  sod[OPT_GAMMA_SCALAR_G].cap |= SANE_CAP_INACTIVE;
                  sod[OPT_GAMMA_SCALAR_B].cap |= SANE_CAP_INACTIVE;
                }
              else
                {
                  sod[OPT_GAMMA_SCALAR].cap |= SANE_CAP_INACTIVE;
                  sod[OPT_GAMMA_SCALAR_R].cap &= ~SANE_CAP_INACTIVE;
                  sod[OPT_GAMMA_SCALAR_G].cap &= ~SANE_CAP_INACTIVE;
                  sod[OPT_GAMMA_SCALAR_B].cap &= ~SANE_CAP_INACTIVE;
                }
            }
          break;

        case MS_MODE_GRAY:
          sod[OPT_GAMMA_MODE].cap &= ~SANE_CAP_INACTIVE;
          sod[OPT_GAMMA_BIND].cap |= SANE_CAP_INACTIVE;
          sod[OPT_GAMMA_SCALAR_R].cap |= SANE_CAP_INACTIVE;
          sod[OPT_GAMMA_SCALAR_G].cap |= SANE_CAP_INACTIVE;
          sod[OPT_GAMMA_SCALAR_B].cap |= SANE_CAP_INACTIVE;
          if ( strcmp(val[OPT_GAMMA_MODE].s, MD_GAMMAMODE_LINEAR) == 0 )
              sod[OPT_GAMMA_SCALAR].cap |= SANE_CAP_INACTIVE;
          else
              sod[OPT_GAMMA_SCALAR].cap &= ~SANE_CAP_INACTIVE;

          break;
      
        case MS_MODE_HALFTONE:
        case MS_MODE_LINEART:
          val[OPT_GAMMA_MODE].s = MD_GAMMAMODE_LINEAR;         
          sod[OPT_GAMMA_MODE].cap |= SANE_CAP_INACTIVE;
          sod[OPT_GAMMA_BIND].cap |= SANE_CAP_INACTIVE;
          sod[OPT_GAMMA_SCALAR].cap |= SANE_CAP_INACTIVE;
          sod[OPT_GAMMA_SCALAR_R].cap |= SANE_CAP_INACTIVE;
          sod[OPT_GAMMA_SCALAR_G].cap |= SANE_CAP_INACTIVE;
          sod[OPT_GAMMA_SCALAR_B].cap |= SANE_CAP_INACTIVE;
          break;
        
        default:
          DBG(1, "restore_gamma_options: unknown mode %d\n", mode);
          break;
    }
    return SANE_STATUS_GOOD;
}
              

/*---------- scsi_wait_for_ready() -------------------------------------------*/

static SANE_Status
scsi_wait_for_ready(Microtek2_Scanner *ms)
{
    int retry = 60;
    SANE_Int status;


    DBG(30, "scsi_wait_for_ready: ms=%p\n", ms);

    while ( retry-- > 0 ) 
      {
        status = scsi_read_image_status(ms);
        if  (status == SANE_STATUS_DEVICE_BUSY ) 
          { 
            sleep(1);
            continue;
          }
        if ( status == SANE_STATUS_GOOD )
            return status;

        /* status != GOOD && != BUSY */
        DBG(1, "scsi_wait_for_ready: '%s'\n", sane_strstatus(status));
        return status;
      }
    
    /* BUSY after n retries */   
    DBG(1, "scsi_wait_for_ready: '%s'\n", sane_strstatus(status));
    return status;
}


/*---------- scsi_read_gamma() -----------------------------------------------*/

/* currently not used */
/*
static SANE_Status
scsi_read_gamma(Microtek2_Scanner *ms, int color)
{
    u_int8_t readgamma[RG_CMD_L];
    u_int8_t result[3072];
    size_t size;
    SANE_Bool endiantype;
    SANE_Status status;

    RG_CMD(readgamma);
    ENDIAN_TYPE(endiantype);   
    RG_PCORMAC(readgamma, endiantype);
    RG_COLOR(readgamma, color);
    RG_WORD(readgamma, ( ms->dev->lut_entry_size == 1 ) ? 0 : 1);
    RG_TRANSFERLENGTH(readgamma, (color == 3 ) ? 3072 : 1024);

    dump_area(readgamma, 10, "ReadGamma");

    size = sizeof(result);
    status = sanei_scsi_cmd(ms->sfd, readgamma, sizeof(readgamma), 
                            result, &size);
    if ( status != SANE_STATUS_GOOD ) {
        DBG(1, "scsi_read_gamma: (L,R) read_gamma failed: status '%s'\n",
                sane_strstatus(status));
        return status;
    }

    dump_area(result, 3072, "Result");

    return SANE_STATUS_GOOD;
}
*/


/*---------- scsi_send_gamma() -----------------------------------------------*/

static SANE_Status
scsi_send_gamma(int fd, u_int8_t *cmdpos, int lut_size, 
                int word_size, int color)
{
    SANE_Bool endiantype;
    SANE_Status status;

    DBG(30, "scsi_send_gamma: pos=%p, word_size=%d, lut_size=%d\n", 
             cmdpos, word_size, lut_size);

    SG_CMD(cmdpos);
    ENDIAN_TYPE(endiantype);   
    SG_PCORMAC(cmdpos, endiantype);
    SG_COLOR(cmdpos, color);
    SG_WORD(cmdpos, word_size - 1);
    SG_TRANSFERLENGTH(cmdpos, lut_size * word_size);

    if ( ms_dump >= 2 )
        dump_area2(cmdpos, SG_CMD_L, "sendgammacmd");
    if ( ms_dump >= 3 )
        dump_area2(cmdpos + SG_CMD_L, lut_size * word_size, "sendgammadata");

    status = sanei_scsi_cmd(fd, cmdpos, lut_size * word_size + SG_CMD_L,
                            NULL, 0);
    if ( status != SANE_STATUS_GOOD ) 
        DBG(1, "scsi_send_gamma: '%s'\n", sane_strstatus(status));

    return status;
}


/*---------- scsi_inquiry() --------------------------------------------------*/

static SANE_Status
scsi_inquiry(Microtek2_Info *mi, char *device)
{
    u_int8_t inquiry[INQ_CMD_L];
    u_int8_t result[INQ_ALLOCATION];
    size_t size;
    int sfd;
    int status;


    DBG(30, "scsi_inquiry: mi=%p, device='%s'\n", mi, device);
    
    INQ_CMD(inquiry);
    status = sanei_scsi_open(device, &sfd, scsi_sense_handler, 0);
    if ( status != SANE_STATUS_GOOD ) {
        DBG(1, "scsi_inquiry: open '%s'\n", sane_strstatus(status));
        return status;
    }

    size = sizeof(result);   
    if (ms_dump >= 2 )
        dump_area2(inquiry, sizeof(inquiry), "inquiry");
    status = sanei_scsi_cmd(sfd, inquiry, sizeof(inquiry), result, &size);
                            
    if ( status != SANE_STATUS_GOOD ) {
        DBG(1, "scsi_inquiry: cmd '%s'\n", sane_strstatus(status));
        sanei_scsi_close(sfd);
        return status;
    }
    sanei_scsi_close(sfd); 

    if (ms_dump >= 2 )
        dump_area2((u_int8_t *) result, sizeof(result), "inquiryresult");

    /* copy results */
    INQ_QUAL(mi->device_qualifier, result);
    INQ_DEVT(mi->device_type, result);
    INQ_VENDOR(mi->vendor, result);
    INQ_MODEL(mi->model, result);
    INQ_REV(mi->revision, result);
    INQ_MODELCODE(mi->model_code, result);

    return SANE_STATUS_GOOD;
}


/*---------- scsi_read_attributes() ------------------------------------------*/

static SANE_Status
scsi_read_attributes(Microtek2_Info *mi, char *device)
{
    u_int8_t readattributes[RSA_CMD_L];
    u_int8_t result[RSA_TRANSFERLENGTH];
    size_t size;
    int sfd;
    int status;
    

    DBG(30, "scsi_read_attributes: mi=%p, device='%s'\n", mi, device);
   
    RSA_CMD(readattributes); 
    status = sanei_scsi_open(device, &sfd, scsi_sense_handler, 0);
    if ( status != SANE_STATUS_GOOD ) 
      {
        DBG(15, "scsi_read_attributes: open '%s'\n", sane_strstatus(status));
        return status;
      }

    size = sizeof(result);   
    if (ms_dump >= 2 )
        dump_area2(readattributes, sizeof(readattributes), "scannerattributes");
                   
    status = sanei_scsi_cmd(sfd, readattributes,
                            sizeof(readattributes), result, &size);
    if ( status != SANE_STATUS_GOOD ) 
      {
        DBG(1, "scsi_read_attributes: cmd '%s'\n", sane_strstatus(status));
        sanei_scsi_close(sfd);
        return status;
      }
    sanei_scsi_close(sfd); 
    
    /* copy all the stuff into the info structure */
    RSA_COLOR(mi->color, result);
    RSA_ONEPASS(mi->onepass, result);
    RSA_SCANNERTYPE(mi->scanner_type, result);
    RSA_FEPROM(mi->feprom, result);
    RSA_DATAFORMAT(mi->data_format, result);
    RSA_COLORSEQUENCE(mi->color_sequence, result);
    RSA_CCDGAP(mi->ccd_gap, result);
    RSA_MAX_XRESOLUTION(mi->max_xresolution, result);
    RSA_MAX_YRESOLUTION(mi->max_yresolution, result);
    RSA_GEOWIDTH(mi->geo_width, result);
    RSA_GEOHEIGHT(mi->geo_height, result);
    RSA_OPTRESOLUTION(mi->opt_resolution, result);
    RSA_DEPTH(mi->depth, result);
    RSA_SCANMODE(mi->scanmode, result);
    RSA_CCDPIXELS(mi->ccd_pixels, result);
    RSA_LUTCAP(mi->lut_cap, result);
    RSA_DNLDPTRN(mi->has_dnldptrn, result);
    RSA_GRAINSLCT(mi->grain_slct, result);
    RSA_SUPPOPT(mi->option_device, result);
    RSA_CALIBWHITE(mi->calib_white, result);
    RSA_CALIBSPACE(mi->calib_space, result);
    RSA_NLENS(mi->nlens, result);
    RSA_NWINDOWS(mi->nwindows, result);
    RSA_SHTRNSFEREQU(mi->shtrnsferequ, result);
    RSA_SCNBTTN(mi->scnbuttn, result);
    RSA_BUFTYPE(mi->buftype, result);
    if (ms_dump >= 2 )
        dump_area2((u_int8_t *) result, sizeof(result), 
                   "scannerattributesresults");
    if ( ms_dump >= 1 && ms_dump_clear )
        dump_attributes(mi);
    
    return SANE_STATUS_GOOD;
}


/*---------- scsi_set_window() -----------------------------------------------*/

static SANE_Status
scsi_set_window(Microtek2_Scanner *ms, SANE_Int n) {   /* n windows, not yet */
                                                       /* implemented */
    u_int8_t setwindow[SW_CMD_L + SW_HEADER_L + n * SW_BODY_L];
    SANE_Int status;


    DBG(30, "scsi_set_window: ms=%p, wnd=%d\n", ms, n);

    memset(setwindow, 0, sizeof(setwindow));
 
    SW_CMD(setwindow);
    SW_PARAM_LENGTH(setwindow, SW_HEADER_L+n*SW_BODY_L);
    SW_WNDDESCLEN(setwindow + SW_HEADER_P, SW_WNDDESCVAL);

    #define POS  (setwindow + SW_BODY_P(n-1))

    SW_WNDID(POS, n-1);
    SW_XRESDPI(POS, ms->x_resolution_dpi);
    SW_YRESDPI(POS, ms->y_resolution_dpi);
    SW_XPOSTL(POS, ms->x1_dots);
    SW_YPOSTL(POS, ms->y1_dots);
    SW_WNDWIDTH(POS, ms->width_dots);
    SW_WNDHEIGHT(POS, ms->height_dots);
    SW_THRESHOLD(POS, ms->threshold);
    SW_IMGCOMP(POS, ms->mode);
    SW_BITSPERPIXEL(POS, ms->depth);
    SW_EXTHT(POS, ms->use_external_ht);
    SW_INTHTINDEX(POS, ms->internal_ht_index);
    SW_RIF(POS, 1); 
    SW_LENS(POS, 0);                                  /* ???? */
    SW_INFINITE(POS, 0);
    SW_STAY(POS, 0);
    SW_RAWDAT(POS, 0);
    SW_QUALITY(POS, ms->quality);
    SW_FASTSCAN(POS, ms->fastscan);
    /* SW_MEDIA(POS, ms->scanner_type); */
    SW_MEDIA(POS, 0);
    SW_BRIGHTNESS_M(POS, ms->brightness_m);
    SW_CONTRAST_M(POS, ms->contrast_m);
    SW_EXPOSURE_M(POS, ms->exposure_m);
    SW_SHADOW_M(POS, ms->shadow_m);
    SW_MIDTONE_M(POS, ms->midtone_m);
    SW_HIGHLIGHT_M(POS, ms->highlight_m);
    /* the following properties are only referenced if it's a color scan */
    /* but I guess they don't matter at a gray scan */
    SW_BRIGHTNESS_R(POS, ms->brightness_r);
    SW_CONTRAST_R(POS, ms->contrast_r);
    SW_EXPOSURE_R(POS, ms->exposure_r);
    SW_SHADOW_R(POS, ms->shadow_r);
    SW_MIDTONE_R(POS, ms->midtone_r);
    SW_HIGHLIGHT_R(POS, ms->highlight_r);
    SW_BRIGHTNESS_G(POS, ms->brightness_g);
    SW_CONTRAST_G(POS, ms->contrast_g);
    SW_EXPOSURE_G(POS, ms->exposure_g);
    SW_SHADOW_G(POS, ms->shadow_g);
    SW_MIDTONE_G(POS, ms->midtone_g);
    SW_HIGHLIGHT_G(POS, ms->highlight_g);
    SW_BRIGHTNESS_B(POS, ms->brightness_b);
    SW_CONTRAST_B(POS, ms->contrast_b);
    SW_EXPOSURE_B(POS, ms->exposure_b);
    SW_SHADOW_B(POS, ms->shadow_b);
    SW_MIDTONE_B(POS, ms->midtone_b);
    SW_HIGHLIGHT_B(POS, ms->highlight_b);

    if ( ms_dump >= 2 )
      {
        dump_area2(setwindow, 10, "setwindowcmd");
        dump_area2(setwindow + 10 ,8 , "setwindowheader");
        dump_area2(setwindow + 18 ,61 , "setwindowbody");
      }

    status = sanei_scsi_cmd(ms->sfd, setwindow, sizeof(setwindow), NULL, 0);
    if ( status != SANE_STATUS_GOOD ) 
        DBG(1, "scsi_set_window: '%s'\n", sane_strstatus(status));
               
    return status;
}


/*---------- scsi_read_image_info() ------------------------------------------*/

static SANE_Status
scsi_read_image_info(Microtek2_Scanner *ms)
{
    u_int8_t rdimginfo[RII_CMD_L];
    u_int8_t result[RII_RESULT_L];
    size_t size;
    SANE_Status status;

    DBG(30, "scsi_read_image_info: ms=%p\n", ms);
    
    RII_CMD(rdimginfo);

    if ( ms_dump >= 2)
        dump_area2(rdimginfo, RII_CMD_L, "readimageinfo");
    
    size = sizeof(result);
    status = sanei_scsi_cmd(ms->sfd, rdimginfo, sizeof(rdimginfo), 
                            result, &size);

    if ( status != SANE_STATUS_GOOD ) 
      {
        DBG(1, "scsi_read_image_info: '%s'\n", sane_strstatus(status));
        return status;
      }

    if ( ms_dump >= 2)
        dump_area2(result, size, "readimageinforesult");
    ms->ppl = RII_WIDTHPIXEL(result);
    ms->bpl = RII_WIDTHBYTES(result);
    ms->src_remaining_lines = RII_HEIGHTLINES(result);
    ms->remaining_bytes = RII_REMAINBYTES(result);
    DBG(30, "scsi_read_image_info: ppl=%d, bpl=%d, lines=%d, remain=%d\n",
             ms->ppl, ms->bpl, ms->src_remaining_lines, ms->remaining_bytes);

    return SANE_STATUS_GOOD;
}


/*---------- scsi_read_image() -----------------------------------------------*/

static SANE_Status
scsi_read_image(Microtek2_Scanner *ms)
{
    u_int8_t readimage[RI_CMD_L];
    SANE_Bool endiantype;
    SANE_Status status;
    size_t size;


    DBG(30, "scsi_read_image:  ms=%p\n", ms);

    RI_CMD(readimage);
    ENDIAN_TYPE(endiantype);   
    RI_PCORMAC(readimage, endiantype);
    if ( ms->do_three_pass )    /* is this if clause necessary ??? */
        RI_COLOR(readimage, ms->current_color);
    RI_TRANSFERLENGTH(readimage, ms->transfer_length);
    DBG(30, "scsi_read_image: transferlength=%d\n", ms->transfer_length);
   
    if ( ms_dump >= 2 ) 
        dump_area2(readimage, RI_CMD_L, "readimagecmd");
    
    size = sanei_scsi_max_request_size;
    status = sanei_scsi_cmd(ms->sfd, readimage, sizeof(readimage), 
                            ms->src_buffer, &size);
    if ( status != SANE_STATUS_GOOD ) 
        DBG(1, "scsi_read_image: '%s'\n", sane_strstatus(status));
   
    if ( ms_dump >= 3 ) 
        dump_area2(ms->src_buffer,ms->transfer_length, "readimageresult");
                
    return status;
}


/*---------- scsi_read_status() ----------------------------------------------*/

static SANE_Status
scsi_read_image_status(Microtek2_Scanner *ms)
{
    u_int8_t rdimgstatus[RIS_CMD_L];
    SANE_Status status;
    SANE_Bool endian_type;
    
    DBG(30, "scsi_read_image_status: ms=%p\n", ms);
   
    ENDIAN_TYPE(endian_type);
    RIS_CMD(rdimgstatus);
    RIS_PCORMAC(rdimgstatus, endian_type);
    if ( ms->do_three_pass )
        RIS_COLOR(rdimgstatus, ms->current_color);
    else
        /* test */
        RIS_COLOR(rdimgstatus, 3);

    if ( ms_dump >= 2 )
        dump_area2(rdimgstatus, sizeof(rdimgstatus), "readimagestatus");

    /* dump_area(rdimgstatus, 10, "ReadImageStatus"); */ /* test */
    status = sanei_scsi_cmd(ms->sfd, rdimgstatus, sizeof(rdimgstatus), 0, NULL);
    if ( status != SANE_STATUS_GOOD ) 
        DBG(1, "scsi_read_image_status: '%s'\n", sane_strstatus(status));
   
    return status;
}


/*---------- scsi_request_sense() --------------------------------------------*/

/* currently not used */
/* static SANE_Status
scsi_request_sense(Microtek2_Scanner *ms)
{
    u_int8_t requestsense[RQS_CMD_L];
    u_int8_t buffer[100];
    SANE_Status status;
    int size;
    int asl;
    int as_info_length;

    DBG(30, "scsi_request_sense: ms=%p\n", ms);
    
    RQS_CMD(requestsense);
    RQS_ALLOCLENGTH(requestsense, 100);

    size = sizeof(buffer);
    status = sanei_scsi_cmd(ms->sfd,  requestsense, sizeof(requestsense), 
                            buffer, &size);
    
    if ( status != SANE_STATUS_GOOD ) 
      {
        DBG(1, "scsi_request_sense: '%s'\n", sane_strstatus(status));
        return status;
      }
           
    if ( ms_dump >= 2 )
        dump_area2(buffer, size, "requestsenseresult");

    dump_area(buffer, RQS_LENGTH(buffer), "RequestSense");
    asl = RQS_ASL(buffer);
    if ( (as_info_length = RQS_ASINFOLENGTH(buffer)) > 0 )
        DBG(25, "scsi_request_sense: info '%.*s'\n",
                as_info_length, RQS_ASINFO(buffer));
    
    return SANE_STATUS_GOOD;
}
*/


/*---------- scsi_sense_handler() --------------------------------------------*/

static SANE_Status
scsi_sense_handler (int fd, u_char *sense, void *arg)
{
    int as_info_length;
    u_int8_t sense_key;
    u_int8_t asl;
    u_int8_t asc;
    u_int8_t ascq;
    
    
    DBG(30, "scsi_sense_handler: fd=%d, sense=%p arg=%p\n",fd, sense, arg);
           
    dump_area(sense, RQS_LENGTH(sense), "SenseBuffer");

    sense_key = RQS_SENSEKEY(sense);
    asl = RQS_ASL(sense);
    asc = RQS_ASC(sense);
    ascq = RQS_ASCQ(sense);
    
    if ( (as_info_length = RQS_ASINFOLENGTH(sense)) > 0 )
        DBG(30,"scsi_sense_handler: info: '%*s'\n",
                as_info_length, RQS_ASINFO(sense));
        
    switch ( sense_key ) 
      {
        case RQS_SENSEKEY_NOSENSE:
          return SANE_STATUS_GOOD;

        case RQS_SENSEKEY_HWERR:
          if ( asc == 0x4a && ascq == 0x00 )
              DBG(5, "scsi_sense_handler: Command phase error\n");
          else if ( asc == 0x4b && ascq == 0x00 )
              DBG(5, "scsi_sense_handler: Data phase error\n");
          else if ( asc == 0x40 ) 
            {
              switch ( ascq ) 
                {
                  case RQS_ASCQ_CPUERR:
                    DBG(5, "scsi_sense_handler: CPU error\n");
                    break;
                  case RQS_ASCQ_SRAMERR:
                    DBG(5, "scsi_sense_handler: SRAM error\n");
                    break;
                  case RQS_ASCQ_DRAMERR:
                    DBG(5, "scsi_sense_handler: DRAM error\n");
                    break;
                  case RQS_ASCQ_DCOFF:
                    DBG(5, "scsi_sense_handler: DC Offset error\n");
                    break;
                  case RQS_ASCQ_GAIN:
                    DBG(5, "scsi_sense_handler: Gain error\n");
                    break;
                  case RQS_ASCQ_POS:
                    DBG(5, "scsi_sense_handler: Pos. error\n");
                    break;
                  default:
                    DBG(5, "scsi_sense_handler: Unknown combination of ASC"
                           " (0x%02x) and ASCQ (0x%02x)\n", asc, ascq);
                    break;
                }
            }   
          else if ( asc == 0x00  && ascq == 0x05) 
              DBG(5, "scsi_sense_handler: End of data detected\n");
          else if ( asc == 0x60 && ascq == 0x00 ) 
              DBG(5, "scsi_sense_handler: Lamp failure\n");
          else if ( asc == 0x53 && ascq == 0x00 ) 
            {
              DBG(5, "scsi_sense_handler: ADF paper jam or no paper\n");
              return SANE_STATUS_NO_DOCS;     /* not STATUS_INVAL */
            }
          else if ( asc == 0x3a && ascq == 0x00 )
              DBG(5, "scsi_sense_handler: Media (ADF or TMA) not available\n");
          else if ( asc == 0x03 && ascq == 0x00 ) 
              DBG(5, "scsi_sense_handler: Peripheral device write fault\n");
          else if ( asc == 0x80 && ascq == 0x00 ) 
              DBG(5, "scsi_sense_handler: Target abort scan\n");
          else
              DBG(5, "scsi_sense_handler: Unknown combination of SENSE KEY "
                     "(0x%02x), ASC (0x%02x) and ASCQ (0x%02x)\n", 
                      sense_key, asc, ascq);

          return SANE_STATUS_IO_ERROR;   

        case RQS_SENSEKEY_ILLEGAL:
          if ( asc == 0x2c && ascq == 0x00 )
              DBG(5, "scsi_sense_handler: Command sequence error\n");
          else if ( asc == 0x3d  && ascq == 0x00) 
              DBG(5, "scsi_sense_handler: Invalid bit in IDENTIFY\n");
          else if ( asc == 0x2c && ascq == 0x02 ) 
/* Ok */      DBG(5, "scsi_sense_handler: Invalid comb. of windows specfied\n");
          else if ( asc == 0x20 && ascq == 0x00 ) 
/* Ok */      DBG(5, "scsi_sense_handler: Invalid command opcode\n");
          else if ( asc == 0x24 && ascq == 0x00 ) 
/* Ok */      DBG(5, "scsi_sense_handler: Invalid field in CDB\n");
          else if ( asc == 0x26 && ascq == 0x00 )
              DBG(5, "scsi_sense_handler: Invalid field in the param list\n");
          else if ( asc == 0x49 && ascq == 0x00 ) 
              DBG(5, "scsi_sense_handler: Invalid message error\n");
          else if ( asc == 0x25 && ascq == 0x00 ) 
              DBG(5, "scsi_sense_handler: Unsupported logic. unit\n");
          else if ( asc == 0x00 && ascq == 0x00 ) 
              DBG(5, "scsi_sense_handler:  No additional sense information\n");
/* Ok */  else if ( asc == 0x1a && ascq == 0x00 ) 
              DBG(5, "scsi_sense_handler: Parameter list length error\n");
          else if ( asc == 0x26 && ascq == 0x02 ) 
              DBG(5, "scsi_sense_handler: Parameter value invalid\n");
          else if ( asc == 0x2c && ascq == 0x01 ) 
              DBG(5, "scsi_sense_handler: Too many windows\n");
          else
              DBG(5, "scsi_sense_handler: Unknown combination of SENSE KEY "
                     "(0x%02x), ASC (0x%02x) and ASCQ (0x%02x)\n", 
                      sense_key, asc, ascq);
             
            return SANE_STATUS_IO_ERROR;   

        case RQS_SENSEKEY_VENDOR:
          DBG(5, "scsi_sense_handler: Vendor specific SENSE KEY (0x%02x), "
                 "ASC (0x%02x) and ASCQ (0x%02x)\n", sense_key, asc, ascq);
             
          return SANE_STATUS_IO_ERROR;   

        default:
           DBG(5, "scsi_sense_handler: (L,R): Unknown sense key (0x%02x)\n",
                   sense_key);
           return SANE_STATUS_IO_ERROR;
    }

    return SANE_STATUS_GOOD;
}


/*---------- scsi_test_unit_ready() ------------------------------------------*/

static SANE_Status
scsi_test_unit_ready(Microtek2_Device *md)
{
    u_int8_t tur[TUR_CMD_L];
    int sfd;
    int status;


    DBG(30, "scsi_test_unit_ready: md=%s\n", md->name);
   
    TUR_CMD(tur); 
    status = sanei_scsi_open(md->name, &sfd, scsi_sense_handler, 0);
    if ( status != SANE_STATUS_GOOD )
      {
	DBG(1, "scsi_test_unit_ready: open '%s'\n", sane_strstatus(status));
	return status;
      }

    if ( ms_dump >= 2 )
        dump_area2(tur, sizeof(tur), "testunitready");

    status = sanei_scsi_cmd(sfd, tur, sizeof(tur), NULL, 0);
    if ( status != SANE_STATUS_GOOD )
        DBG(1, "scsi_test_unit_ready: cmd '%s'\n", sane_strstatus(status));
      
    sanei_scsi_close(sfd); 
    return status;
}
